from policyengine_core.model_api import *
from policyengine_core import periods


def str_to_instant(s):
    return periods.Instant(tuple(map(lambda s: int(s), s.split("-"))))


def backdate_parameters(
    root: str = None, first_instant: str = "2021-01-01"
) -> Reform:
    first_instant = str_to_instant(first_instant)
    node = root
    for param in node.get_descendants():
        if hasattr(param, "values_list"):
            earliest = param.values_list[-1]
            earliest_value = earliest.value
            earliest_instant = str_to_instant(earliest.instant_str)
            if first_instant < earliest_instant:
                num_days = (earliest_instant.date - first_instant.date).days
                param.update(
                    period=periods.Period(("day", first_instant, num_days)),
                    value=earliest_value,
                )
    return root



from policyengine_us.model_api import *


class ms_income_tax_before_credits_indiv(Variable):
    value_type = float
    entity = Person
    label = "Mississippi income tax before credits when married couples file separately"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MS

    def formula(person, period, parameters):
        income = person("ms_taxable_income_indiv", period)
        rate = parameters(period).gov.states.ms.tax.income.rate
        return rate.calc(income)



from policyengine_us.model_api import *


class md_ctc_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Eligible for the Maryland Child Tax Credit"
    definition_period = YEAR
    reference = [
        "https://casetext.com/statute/code-of-maryland/article-tax-general/title-10-income-tax/subtitle-7-income-tax-credits/section-10-751-effective-until-712026-tax-credit-for-qualified-child",
        "https://law.justia.com/codes/maryland/2022/tax-general/title-10/subtitle-7/section-10-751/",
        "https://mgaleg.maryland.gov/2025RS/Chapters_noln/CH_604_hb0352e.pdf#page=169",  # Maryland House Bill 352 - Budget Reconciliation and Financing Act of 2025
    ]
    defined_for = StateCode.MD

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.md.tax.income.credits.ctc
        agi = tax_unit("adjusted_gross_income", period)

        # When phase-out applies, eligibility is based on having eligible children
        # (phase-out is handled in the credit amount calculation)
        if p.phase_out.applies:
            person = tax_unit.members
            dependent = person("is_tax_unit_dependent", period)
            disabled = person("is_disabled", period)
            age_limit = where(
                disabled, p.age_threshold.disabled, p.age_threshold.main
            )
            meets_age_limit = person("age", period) < age_limit
            eligible = dependent & meets_age_limit
            return tax_unit.sum(eligible) > 0

        # When phase-out does not apply, use old logic (hard adjusted gross income cutoff)
        return agi <= p.agi_cap



from policyengine_us.model_api import *


class basic_standard_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Basic standard deduction"
    definition_period = YEAR
    unit = USD
    reference = "https://www.law.cornell.edu/uscode/text/26/63#c_2"

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.irs.deductions.standard
        filing_status = tax_unit("filing_status", period)
        separate_filer_itemizes = tax_unit("separate_filer_itemizes", period)
        dependent_elsewhere = tax_unit("head_is_dependent_elsewhere", period)
        standard_deduction = p.amount[filing_status]
        standard_deduction_if_dependent = min_(
            standard_deduction,
            max_(
                p.dependent.additional_earned_income
                + tax_unit("tax_unit_earned_income", period),
                p.dependent.amount,
            ),
        )
        return select(
            [
                separate_filer_itemizes,
                dependent_elsewhere,
            ],
            [
                0,
                standard_deduction_if_dependent,
            ],
            default=standard_deduction,
        )



from policyengine_us.model_api import *


class al_standard_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Alabama standard deduction"
    unit = USD
    # The Code of Alabama 1975 Section 40-18-15 (b)(4).
    documentation = "https://alisondb.legislature.state.al.us/alison/CodeOfAlabama/1975/Coatoc.htm"
    definition_period = YEAR
    defined_for = StateCode.AL

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.al.tax.income.deductions.standard
        filing_status = tax_unit("filing_status", period)
        base_amount = p.amount.max[filing_status]
        increment = p.phase_out.increment[filing_status]
        min_amount = p.amount.min[filing_status]
        rate = p.phase_out.rate[filing_status]
        threshold = p.phase_out.threshold[filing_status]
        al_agi = tax_unit("al_agi", period)
        # No "or fraction thereof" clause, so use integer (floor) division rather than ceiling.
        excess_income = max_(0, al_agi - threshold)
        increments = excess_income // increment
        reduction = increments * rate

        return max_(base_amount - reduction, min_amount)



from policyengine_us.model_api import *


class de_income_tax_if_claiming_non_refundable_eitc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Delaware tax liability if claiming non-refundable Delaware EITC"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.DE

    def formula(tax_unit, period, parameters):
        simulation = tax_unit.simulation
        non_refundable_branch = simulation.get_branch("de_non_refundable_eitc")
        non_refundable_branch.set_input(
            "de_claims_refundable_eitc",
            period,
            np.zeros((tax_unit.count,), dtype=bool),
        )
        return non_refundable_branch.calculate("de_income_tax", period)



from policyengine_us.model_api import *


class nm_itemized_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "New Mexico itemized deductions"
    unit = USD
    definition_period = YEAR
    reference = "https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf"
    defined_for = StateCode.NM

    adds = ["total_itemized_taxable_income_deductions"]



from policyengine_us.model_api import *


class hi_taxable_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "Hawaii taxable income"
    defined_for = StateCode.HI
    unit = USD
    definition_period = YEAR
    # Hawaii Instructions for Form N-11 Rev 2022 p20
    reference = "https://files.hawaii.gov/tax/forms/2022/n11ins.pdf"

    def formula(tax_unit, period, parameters):
        hi_agi = tax_unit("hi_agi", period)
        hi_deductions = tax_unit("hi_deductions", period)
        hi_exemptions = tax_unit("hi_exemptions", period)

        return max_(0, hi_agi - hi_deductions - hi_exemptions)



from policyengine_us.model_api import *


class vt_taxable_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "Vermont taxable income"
    unit = USD
    documentation = "VT AGI less taxable income deductions and exemptions"
    definition_period = YEAR
    reference = (
        "https://tax.vermont.gov/sites/tax/files/documents/IN-111-2022.pdf"
    )
    defined_for = StateCode.VT

    def formula(tax_unit, period, parameters):
        agi = tax_unit("vt_agi", period)
        deductions = tax_unit("vt_standard_deduction", period)
        exemptions = tax_unit("vt_personal_exemptions", period)
        return max_(0, agi - deductions - exemptions)



from policyengine_us.model_api import *
from policyengine_core.periods import period as period_


def create_dc_property_tax_credit() -> Reform:
    class dc_ptc(Variable):
        value_type = float
        entity = TaxUnit
        label = "DC property tax credit"
        unit = USD
        definition_period = YEAR
        reference = (
            "https://otr.cfo.dc.gov/sites/default/files/dc/sites/otr/publication/attachments/52926_D-40_12.21.21_Final_Rev011122.pdf#page=49"
            "https://otr.cfo.dc.gov/sites/default/files/dc/sites/otr/publication/attachments/2022_D-40_Booklet_Final_blk_01_23_23_Ordc.pdf#page=47"
        )
        defined_for = "takes_up_dc_ptc"

        def formula(tax_unit, period, parameters):
            p = parameters(period).gov.contrib.states.dc.property_tax
            filing_status = tax_unit("filing_status", period)
            amount = p.amount[filing_status]
            # Income based eligibility criteria does not apply when the credit
            # is phased out.
            eligible = tax_unit("dc_ptc_eligible", period)
            if p.phase_out.applies:
                income_limit = tax_unit("dc_ptc_income_limit", period)
                income = tax_unit("adjusted_gross_income", period)
                income_excess = max_(0, income - income_limit)
                return max_(0, amount - p.phase_out.rate * income_excess)
            return amount * eligible

    class dc_ptc_eligible(Variable):
        value_type = bool
        entity = TaxUnit
        label = "DC property tax credit eligible"
        definition_period = YEAR
        reference = (
            "https://otr.cfo.dc.gov/sites/default/files/dc/sites/otr/publication/attachments/52926_D-40_12.21.21_Final_Rev011122.pdf#page=49"
            "https://otr.cfo.dc.gov/sites/default/files/dc/sites/otr/publication/attachments/2022_D-40_Booklet_Final_blk_01_23_23_Ordc.pdf#page=47"
        )
        defined_for = StateCode.DC

        def formula(tax_unit, period, parameters):
            us_agi = tax_unit("adjusted_gross_income", period)
            income_limt = tax_unit("dc_ptc_income_limit", period)
            return us_agi <= income_limt

    class dc_ptc_income_limit(Variable):
        value_type = float
        entity = TaxUnit
        unit = USD
        label = "DC property tax credit income limit"
        definition_period = YEAR
        reference = (
            "https://otr.cfo.dc.gov/sites/default/files/dc/sites/otr/publication/attachments/52926_D-40_12.21.21_Final_Rev011122.pdf#page=49"
            "https://otr.cfo.dc.gov/sites/default/files/dc/sites/otr/publication/attachments/2022_D-40_Booklet_Final_blk_01_23_23_Ordc.pdf#page=47"
        )
        defined_for = StateCode.DC

        def formula(tax_unit, period, parameters):
            elderly_present = tax_unit(
                "dc_ptc_elderly_head_or_spouse_present", period
            )
            filing_status = tax_unit("filing_status", period)
            p = parameters(
                period
            ).gov.contrib.states.dc.property_tax.income_limit
            return where(
                elderly_present,
                p.elderly[filing_status],
                p.non_elderly[filing_status],
            )

    class dc_ptc_elderly_head_or_spouse_present(Variable):
        value_type = bool
        entity = TaxUnit
        label = "DC property tax credit elderly head or spouse present"
        definition_period = YEAR
        reference = (
            "https://otr.cfo.dc.gov/sites/default/files/dc/sites/otr/publication/attachments/52926_D-40_12.21.21_Final_Rev011122.pdf#page=49"
            "https://otr.cfo.dc.gov/sites/default/files/dc/sites/otr/publication/attachments/2022_D-40_Booklet_Final_blk_01_23_23_Ordc.pdf#page=47"
        )
        defined_for = StateCode.DC

        def formula(tax_unit, period, parameters):
            p_dc = parameters(period).gov.states.dc.tax.income.credits
            elderly_age = p_dc.ptc.min_elderly_age
            head_age = tax_unit("age_head", period)
            spouse_age = tax_unit("age_spouse", period)
            return (head_age >= elderly_age) | (spouse_age >= elderly_age)

    class reform(Reform):
        def apply(self):
            self.update_variable(dc_ptc)
            self.update_variable(dc_ptc_eligible)
            self.update_variable(dc_ptc_income_limit)
            self.update_variable(dc_ptc_elderly_head_or_spouse_present)

    return reform


def create_dc_property_tax_credit_reform(
    parameters, period, bypass: bool = False
):
    if bypass:
        return create_dc_property_tax_credit()

    p = parameters.gov.contrib.states.dc.property_tax

    reform_active = False
    current_period = period_(period)

    for i in range(5):
        if p(current_period).in_effect:
            reform_active = True
            break
        current_period = current_period.offset(1, "year")

    if reform_active:
        return create_dc_property_tax_credit()
    else:
        return None


dc_property_tax_credit = create_dc_property_tax_credit_reform(
    None, None, bypass=True
)



from policyengine_us.model_api import *


class me_property_tax_fairness_credit_veterans_cap(Variable):
    value_type = float
    entity = TaxUnit
    unit = USD
    label = "Veterans cap for Maine property tax fairness credit"
    definition_period = YEAR
    defined_for = StateCode.ME
    reference = {
        "title": "2024 SCHEDULE PTFC/STFC Form 1040ME",
        "href": "https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/24_Form%201040ME_Sch%20PTFC_ff_0.pdf#page=2",
    }
    # Only one spouse needs to be a veteran for the filer to get the veteran match

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.me.tax.income.credits.fairness.property_tax
        base_cap = tax_unit("me_property_tax_fairness_credit_base_cap", period)
        # Check if the tax unit head or spouse is a veteran
        person = tax_unit.members
        is_head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        is_veteran = person("is_veteran", period)
        has_veteran_head_or_spouse = tax_unit.any(
            is_head_or_spouse & is_veteran
        )
        return has_veteran_head_or_spouse * base_cap * p.veterans_matched



from policyengine_us.model_api import *


class nm_armed_forces_retirement_pay_exemption_person(Variable):
    value_type = float
    entity = Person
    label = "New Mexico armed forces retirement pay exemption per person "
    unit = USD
    definition_period = YEAR
    reference = "https://casetext.com/statute/new-mexico-statutes-1978/chapter-7-taxation/article-2-income-tax-general-provisions/section-7-2-513-effective-until-112025-exemption-armed-forces-retirement-pay"
    defined_for = StateCode.NM

    def formula(person, period, parameters):
        p = parameters(
            period
        ).gov.states.nm.tax.income.exemptions.armed_forces_retirement_pay
        armed_forces_retirement_pay = person("military_retirement_pay", period)
        capped_amount = min_(armed_forces_retirement_pay, p.cap)
        qualifies = person("is_tax_unit_head_or_spouse", period)
        return qualifies * capped_amount



from policyengine_us.model_api import *


class ky_itemized_deductions_joint(Variable):
    value_type = float
    entity = Person
    label = "Kentucky itemized deductions when married couples file jointly"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://revenue.ky.gov/Forms/Form%20740%20Schedule%20A%202022.pdf"
        "https://law.justia.com/codes/kentucky/2022/chapter-141/section-141-019/"  # (2)(i)
    )
    defined_for = StateCode.KY

    def formula(person, period, parameters):
        # If filing a joint return, only one standard deduction is allowed
        is_head = person("is_tax_unit_head", period)
        itemized_deductions = person.tax_unit(
            "ky_itemized_deductions_unit", period
        )

        return is_head * itemized_deductions



from policyengine_us.model_api import *


class ut_taxpayer_credit_reduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Utah taxpayer credit reduction"
    unit = USD
    documentation = "Form TC-40, line 19"
    definition_period = YEAR
    defined_for = StateCode.UT

    def formula(tax_unit, period, parameters):
        phase_out_income = tax_unit(
            "ut_taxpayer_credit_phase_out_income", period
        )
        phase_out_rate = parameters(
            period
        ).gov.states.ut.tax.income.credits.taxpayer.phase_out.rate
        return phase_out_income * phase_out_rate



from policyengine_us.model_api import *


class ny_qualified_solar_energy_systems_equipment_expenditures(Variable):
    value_type = float
    entity = TaxUnit
    label = "Qualified solar energy systems equipment expenditures in New York"
    unit = USD
    definition_period = YEAR
    reference = "https://www.nysenate.gov/legislation/laws/TAX/606"  # (g)(2)(C)(9)(g-1)(2)
    defined_for = StateCode.NY



from policyengine_us.model_api import *


class pr_low_income_credit_eligible_people(Variable):
    value_type = int
    entity = TaxUnit
    label = "Eligible people for the Puerto Rico low income credit"
    definition_period = YEAR
    reference = "https://casetext.com/statute/laws-of-puerto-rico/title-thirteen-taxation-and-finance/subtitle-17-internal-revenue-code-of-2011/part-ii-income-taxes/chapter-1007-credits-against-tax/subchapter-b-refundable-credits/30212-credit-for-low-income-individuals-older-than-sixty-five-65-years-of-age"
    defined_for = StateCode.PR

    adds = ["pr_low_income_credit_eligible_person"]



from policyengine_us.model_api import *


class ms_national_guard_or_reserve_pay_adjustment(Variable):
    value_type = float
    entity = Person
    label = "Mississippi national guard or reserve pay adjustment"
    unit = USD
    definition_period = YEAR
    reference = [
        "https://www.dor.ms.gov/sites/default/files/Forms/Individual/80100221.pdf#page=12",
        "https://law.justia.com/codes/mississippi/2020/title-27/chapter-7/article-1/section-27-7-18/",  # 4(m)
    ]
    defined_for = StateCode.MS

    def formula(person, period, parameters):
        military_income = person("military_service_income", period)
        p = parameters(
            period
        ).gov.states.ms.tax.income.adjustments.national_guard_or_reserve_pay
        # The tax form instructions confirm the exclusion is capped per taxpayer
        return min_(military_income, p.cap)



from policyengine_us.model_api import *


class is_adult_for_medicaid_fc(Variable):
    value_type = bool
    entity = Person
    label = "Medicaid adult financial criteria"
    definition_period = YEAR

    def formula(person, period, parameters):
        ma = parameters(period).gov.hhs.medicaid.eligibility.categories.adult
        income = person("medicaid_income_level", period)
        state = person.household("state_code_str", period)
        income_limit = ma.income_limit[state]
        return income < income_limit



from policyengine_us.model_api import *


class de_elderly_or_disabled_income_exclusion_eligible_person(Variable):
    value_type = bool
    entity = Person
    label = (
        "Eligible person for the Delaware elderly or disabled income exclusion"
    )
    definition_period = YEAR
    defined_for = StateCode.DE

    def formula(person, period, parameters):
        # First get their filing status.
        filing_status = person.tax_unit(
            "state_filing_status_if_married_filing_separately_on_same_return",
            period,
        )
        p = parameters(
            period
        ).gov.states.de.tax.income.subtractions.exclusions.elderly_or_disabled

        # Get the individual disabled status.\
        disability_eligible = person("is_disabled", period)

        # Get the individual filer's age and eligibility.
        age = person("age", period)
        age_eligible = age >= p.eligibility.age_threshold

        # Get the tax unit income
        earned_income = person("earned_income", period)

        # Determine if filer income is eligible.
        income_threshold = p.eligibility.earned_income_limit[filing_status]
        income_eligible = earned_income < income_threshold

        pre_exclusions_agi = person("de_pre_exclusions_agi", period)
        agi_eligible = (
            pre_exclusions_agi <= p.eligibility.agi_limit[filing_status]
        )

        is_head_or_spouse = person("is_tax_unit_head_or_spouse", period)

        return (
            (age_eligible | disability_eligible)
            & income_eligible
            & agi_eligible
            & is_head_or_spouse
        )



from policyengine_us.model_api import *


class ga_income_tax_before_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "Georgia income tax before refundable credits"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.GA

    def formula(tax_unit, period, parameters):
        tax_before_refundable_credits = tax_unit(
            "ga_income_tax_before_non_refundable_credits", period
        )
        credits = tax_unit("ga_non_refundable_credits", period)
        return max_(0, tax_before_refundable_credits - credits)



from policyengine_us.model_api import *


class il_cta_free_ride_benefit(Variable):
    value_type = float
    entity = Person
    unit = USD
    label = "Illinois Chicago Transit Authority Free Ride Program benefit"
    definition_period = MONTH
    defined_for = "il_cta_free_ride_eligible"
    reference = "https://www.transitchicago.com/fares/"

    adds = ["gov.states.il.rta.cta.monthly_pass_cost.full_fare"]



from policyengine_us.model_api import *


class mt_eitc(Variable):
    value_type = float
    entity = Person
    label = "Montana EITC"
    unit = USD
    definition_period = YEAR
    reference = "https://leg.mt.gov/bills/mca/title_0150/chapter_0300/part_0230/section_0180/0150-0300-0230-0180.html"
    defined_for = StateCode.MT

    def formula(person, period, parameters):
        federal_eitc = person.tax_unit("eitc", period)
        p = parameters(period).gov.states.mt.tax.income.credits.eitc
        # Since the eitc amount can be attributed to either spouse, we will allocate the
        # value to the head
        is_head = person("is_tax_unit_head", period)
        state_eitc = federal_eitc * p.match
        # Separate filers are inleigible for the Montana EITC, this case is currently
        # excluded as we only model separate filing on the same return
        return is_head * state_eitc



from policyengine_us.model_api import *


class ma_ccfa_head_start_partner_and_kindergarten_reimbursement(Variable):
    value_type = float
    entity = Person
    label = "Massachusetts Child Care Financial Assistance (CCFA) head start partner and kindergarten reimbursement amount per child"
    unit = USD
    reference = "https://www.mass.gov/doc/fiscal-year-2025-child-care-financial-assistance-daily-reimbursement-rates/download"
    definition_period = MONTH
    defined_for = StateCode.MA

    def formula(person, period, parameters):
        p = parameters(period).gov.states.ma.eec.ccfa.reimbursement_rates
        region = person.household("ma_ccfa_region", period)
        schedule_type = person("ma_ccfa_schedule_type", period)
        return p.head_start_partner_and_kindergarten[region][schedule_type]



from policyengine_us.model_api import *


class az_taxable_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "Arizona taxable income"
    unit = USD
    documentation = "https://azdor.gov/sites/default/files/2023-03/FORMS_INDIVIDUAL_2022_140Ai.pdf#page=8https://azdor.gov/sites/default/files/2023-03/FORMS_INDIVIDUAL_2022_140Ai.pdf#page=8"
    definition_period = YEAR
    defined_for = StateCode.AZ

    def formula(tax_unit, period, parameters):
        az_agi = tax_unit("az_agi", period)
        deductions = tax_unit("az_deductions", period)
        return max_(0, az_agi - deductions)



from policyengine_us.model_api import *


class vt_low_income_cdcc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Vermont low-income child care and dependent care credit"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://tax.vermont.gov/sites/tax/files/documents/IN-112-2021.pdf#page=2"
        "https://law.justia.com/codes/vermont/2021/title-32/chapter-151/section-5828c/"
    )
    defined_for = "vt_low_income_cdcc_eligible"

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.vt.tax.income.credits.cdcc.low_income
        federal_cdcc = tax_unit("capped_cdcc", period)
        return p.rate * federal_cdcc



from policyengine_us.model_api import *


class refundable_ctc(Variable):
    value_type = float
    entity = TaxUnit
    label = "refundable CTC"
    unit = USD
    documentation = "The portion of the Child Tax Credit that is refundable."
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/24#d"

    def formula(tax_unit, period, parameters):
        # This line corresponds to "the credit which would be allowed under this section [the CTC section]"
        # without regard to this subsection [the refundability section] and the limitation under
        # section 26(a) [the section that limits the amount of the non-refundable CTC to tax liability].
        # This is the full CTC. This is then limited to the maximum refundable amount per child as per the
        # TCJA provision.

        ctc = parameters(period).gov.irs.credits.ctc

        maximum_amount = tax_unit("ctc_refundable_maximum", period)

        total_ctc = tax_unit("ctc", period)

        if ctc.refundable.fully_refundable:
            reduction = tax_unit("ctc_phase_out", period)
            reduced_max_amount = max_(0, maximum_amount - reduction)
            return min_(reduced_max_amount, total_ctc)

        maximum_refundable_ctc = min_(maximum_amount, total_ctc)

        phase_in = tax_unit("ctc_phase_in", period)
        limiting_tax = tax_unit("ctc_limiting_tax_liability", period)
        ctc_capped_by_tax = min_(total_ctc, limiting_tax)
        ctc_capped_by_increased_tax = min_(total_ctc, limiting_tax + phase_in)
        amount_ctc_would_increase = (
            ctc_capped_by_increased_tax - ctc_capped_by_tax
        )
        return min_(maximum_refundable_ctc, amount_ctc_would_increase)



from policyengine_us.model_api import *


class ia_eitc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Iowa earned income tax credit"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://tax.iowa.gov/sites/default/files/2022-01/IA1040%2841-001%29.pdf#page=2"
        "https://tax.iowa.gov/sites/default/files/2023-01/2021%20Expanded%20Instructions_010323.pdf#page=87"
        "https://tax.iowa.gov/sites/default/files/2023-01/2022IA1040%2841001%29.pdf#page=2"
        "https://tax.iowa.gov/sites/default/files/2023-03/2022%20Expanded%20Instructions_022023.pdf#page=87"
    )
    defined_for = StateCode.IA

    def formula(tax_unit, period, parameters):
        federal_eitc = tax_unit("eitc", period)
        p = parameters(period).gov.states.ia.tax.income
        return federal_eitc * p.credits.earned_income.fraction



from policyengine_us.model_api import *


class mn_social_security_subtraction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Minnesota social security subtraction"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.taxformfinder.org/forms/2021/2021-minnesota-form-m1m.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2023-01/m1m_22.pdf"
    )
    defined_for = StateCode.MN

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.mn.tax.income.subtractions.social_security
        filing_status = tax_unit("filing_status", period)
        agi = tax_unit("adjusted_gross_income", period)
        reduction_start = p.reduction.start[filing_status]
        reduction_applies = agi > reduction_start
        us_taxable_oasdi = add(tax_unit, period, ["taxable_social_security"])
        if p.reduction.applies:
            reduction_increment = p.reduction.increment[filing_status]
            reduction_rate = p.reduction.rate
            agi_excess = max_(0, agi - reduction_start)
            reduction_fraction = np.ceil(agi_excess / reduction_increment)
            reduction_rate = min_(reduction_rate * reduction_fraction, 1)
            reduction = reduction_rate * us_taxable_oasdi
            simplified_reduction = max_(us_taxable_oasdi - reduction, 0)
        else:
            simplified_reduction = 0
        # specify parameters
        filing_status = tax_unit("filing_status", period)

        total_benefit_fraction = p.total_benefit_fraction
        income_amount = p.income_amount[filing_status]
        alt_amount = p.alternative_amount[filing_status]
        # calculate subtraction amount (following "Worksheet for line 12")
        # ... US-taxable social security benefits
        # ... alternative benefit subtraction amount
        us_gross_income = add(tax_unit, period, ["irs_gross_income"])
        adj_income = us_gross_income - us_taxable_oasdi
        total_oasdi = add(tax_unit, period, ["social_security"])
        oasdi_amount = total_oasdi * total_benefit_fraction
        tax_exempt_int = add(tax_unit, period, ["tax_exempt_interest_income"])
        sum_income = adj_income + oasdi_amount + tax_exempt_int
        us_ald = tax_unit("above_the_line_deductions", period)
        student_loan_int = add(tax_unit, period, ["student_loan_interest"])
        mn_ald = max_(0, us_ald - student_loan_int)
        income = max_(0, sum_income - mn_ald)
        net_income = max_(0, income - income_amount)
        alt_sub_amt = max_(
            0, alt_amount - (net_income * p.net_income_fraction)
        )
        main_reduction = min_(us_taxable_oasdi, alt_sub_amt)
        if p.reduction.applies:
            return where(
                reduction_applies,
                max_(main_reduction, simplified_reduction),
                us_taxable_oasdi,
            )
        else:
            return max_(main_reduction, simplified_reduction)



from policyengine_us.model_api import *


class va_military_benefit_subtraction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Virginia military benefit subtraction"
    defined_for = StateCode.VA
    unit = USD
    definition_period = YEAR
    reference = (
        "https://law.lis.virginia.gov/vacodefull/title58.1/chapter3/article2/"
    )

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        p = parameters(
            period
        ).gov.states.va.tax.income.subtractions.military_benefit

        subtractable_military_benefit = min_(
            person("military_retirement_pay", period), p.amount
        )
        head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        if p.availability:
            age = person("age", period)
            age_eligible = age >= p.age_threshold
            head_or_spouse_eligible = head_or_spouse & age_eligible
        else:
            head_or_spouse_eligible = head_or_spouse
        return tax_unit.sum(
            subtractable_military_benefit * head_or_spouse_eligible
        )



from policyengine_us.model_api import *


class military_retirement_pay(Variable):
    value_type = float
    entity = Person
    label = "Military retirement pay"
    unit = USD
    definition_period = YEAR
    documentation = "The benefits received under a United States military retirement plan, including survivor benefits."
    reference = "https://militarypay.defense.gov/Pay/Retirement/"



from policyengine_us.model_api import *


class me_property_tax_fairness_credit_cap(Variable):
    value_type = float
    entity = TaxUnit
    unit = USD
    label = "Maine property tax fairness credit cap"
    definition_period = YEAR
    defined_for = StateCode.ME

    adds = [
        "me_property_tax_fairness_credit_base_cap",
        "me_property_tax_fairness_credit_veterans_cap",
    ]



from policyengine_us.model_api import *


class il_ccap_income_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Eligible for Illinois Child Care Assistance Program (CCAP) due to income"
    definition_period = MONTH
    reference = "https://www.dhs.state.il.us/page.aspx?item=118832"
    defined_for = StateCode.IL

    def formula(spm_unit, period, parameters):
        p = parameters(period).gov.states.il.dhs.ccap.income.income_limit
        countable_income = spm_unit("il_ccap_countable_income", period)
        fpg = spm_unit("spm_unit_fpg", period)
        enrolled_in_ccap = spm_unit("il_ccap_enrolled", period)
        fpg_limit = where(
            enrolled_in_ccap,
            p.redetermination_rate,
            p.new_applicants_rate,
        )
        income_limit = fpg_limit * fpg
        return countable_income <= income_limit



from policyengine_us.model_api import *


class il_aabd_grant_amount(Variable):
    value_type = float
    entity = Person
    label = "Illinois Aid to the Aged, Blind or Disabled (AABD) grant amount"
    unit = USD
    definition_period = MONTH
    defined_for = StateCode.IL
    reference = "https://www.dhs.state.il.us/page.aspx?item=15948"

    adds = ["gov.states.il.dhs.aabd.payment.grant_amount"]



from policyengine_us.model_api import *


class va_military_basic_pay_subtraction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Virginia military basic pay subtraction"
    defined_for = StateCode.VA
    unit = USD
    definition_period = YEAR
    reference = (
        "https://law.lis.virginia.gov/vacodefull/title58.1/chapter3/article2/"
    )

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        p = parameters(
            period
        ).gov.states.va.tax.income.subtractions.military_basic_pay
        # Compute subtractable military pay for head and spouse separately.
        military_pay = person("military_basic_pay", period)
        # Subtraction phases in and then out dollar for dollar with respect to military pay, at a given threshold.
        subtractable_military_pay = where(
            military_pay < p.threshold,
            military_pay,
            max_(0, (2 * p.threshold) - military_pay),
        )
        is_head_or_spouse = person("is_tax_unit_head", period) | person(
            "is_tax_unit_spouse", period
        )
        # Sum subtractable military pay for heads and spouses.
        return tax_unit.sum(subtractable_military_pay * is_head_or_spouse)



from policyengine_us.model_api import *


class ma_ccfa_immigration_status_eligible(Variable):
    value_type = bool
    entity = Person
    label = "Immigration status eligible for Massachusetts Child Care Financial Assistance (CCFA)"
    definition_period = MONTH
    defined_for = StateCode.MA
    reference = "https://www.mass.gov/doc/eecs-financial-assistance-policy-guide-february-1-2022/download#page=26"

    def formula(person, period, parameters):
        p = parameters(period).gov.states.ma.eec.ccfa
        immigration_status = person("immigration_status", period)
        immigration_status_str = immigration_status.decode_to_str()
        return np.isin(
            immigration_status_str,
            p.qualified_immigration_statuses,
        )



from policyengine_us.model_api import *


class pell_grant_efc(Variable):
    value_type = float
    entity = Person
    unit = USD
    label = "Pell Grant expected family contribution"
    definition_period = YEAR

    def formula(person, period, parameters):
        head_contribution = person("pell_grant_head_contribution", period)
        dependent_contribution = person(
            "pell_grant_dependent_contribution", period
        )
        head_income = person.tax_unit("pell_grant_primary_income", period)
        formula = person("pell_grant_formula", period).decode_to_str()
        zero_efc_max = parameters(period).gov.ed.pell_grant.efc.automatic_zero
        automatic_zero = head_income <= zero_efc_max
        return select(
            [formula == "A", formula == "B", formula == "C"],
            [
                where(
                    automatic_zero,
                    0,
                    max_(0, head_contribution + dependent_contribution),
                ),
                max_(0, head_contribution),
                where(automatic_zero, 0, max_(0, head_contribution)),
            ],
        )



from policyengine_us.model_api import *


class md_aged_exemption(Variable):
    value_type = float
    entity = TaxUnit
    label = "MD aged exemption"
    unit = USD
    definition_period = YEAR
    reference = "https://govt.westlaw.com/mdc/Document/NF59A76006EA511E8ABBEE50DE853DFF4?viewType=FullText&originationContext=documenttoc&transitionType=CategoryPageItem&contextData=(sc.Default)"
    defined_for = StateCode.MD

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.md.tax.income.exemptions.aged
        aged_head = (tax_unit("age_head", period) >= p.age).astype(int)
        aged_spouse = (tax_unit("age_spouse", period) >= p.age).astype(int)
        return aged_head * p.amount + aged_spouse * p.amount



from policyengine_us.model_api import *


class co_tanf_count_children(Variable):
    value_type = int
    entity = SPMUnit
    label = "Colorado TANF number of children"
    definition_period = YEAR
    defined_for = StateCode.CO
    adds = ["is_child", "is_pregnant"]



from policyengine_us.model_api import *


class ne_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "NE refundable income tax credits"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://revenue.nebraska.gov/files/doc/tax-forms/2021/f_1040n_booklet.pdf"
        "https://revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_5.pdf"
    )
    defined_for = StateCode.NE
    adds = "gov.states.ne.tax.income.credits.refundable"



from policyengine_us.model_api import *


class ia_taxable_income_indiv(Variable):
    value_type = float
    entity = Person
    label = "Iowa taxable income when married couple file separately"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://tax.iowa.gov/sites/default/files/2022-01/IA1040%2841-001%29.pdf"
        "https://tax.iowa.gov/sites/default/files/2023-01/2021%20Expanded%20Instructions_010323.pdf"
        "https://tax.iowa.gov/sites/default/files/2023-01/2022IA1040%2841001%29.pdf"
        "https://tax.iowa.gov/sites/default/files/2023-03/2022%20Expanded%20Instructions_022023.pdf"
    )
    defined_for = StateCode.IA

    def formula(person, period, parameters):
        net_income = person("ia_net_income", period)
        p = parameters(period).gov.states.ia.tax.income
        deductions = [
            f"{ded}_indiv" if ded == "ia_basic_deduction" else ded
            for ded in p.deductions.sources
        ]
        deductions_amount = add(person, period, deductions)
        return max_(0, net_income - deductions_amount)



from policyengine_us.model_api import *


class dc_liheap_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Eligible for the DC LIHEAP"
    definition_period = YEAR
    defined_for = StateCode.DC
    reference = "https://doee.dc.gov/liheap"

    def formula(spm_unit, period, parameters):
        p = parameters(period).gov.hhs.liheap
        state_median_income = spm_unit("hhs_smi", period)
        # The income concept is not clearly defined, assuming IRS gross income
        income = add(spm_unit, period, ["irs_gross_income"])
        smi_limit = state_median_income * p.smi_limit
        return income <= smi_limit



from policyengine_us.model_api import *


class spm_unit_oecd_equiv_net_income(Variable):
    value_type = float
    entity = SPMUnit
    label = "Equivalised income"
    unit = USD
    documentation = "Equivalised net income for the SPM unit under the OECD method (divided by the sqare root of the number of persons in the household)"
    definition_period = YEAR
    reference = (
        "https://www.oecd.org/economy/growth/OECD-Note-EquivalenceScales.pdf"
    )

    def formula(spm_unit, period, parameters):
        number_of_people = spm_unit.nb_persons()
        net_income = spm_unit("spm_unit_net_income", period)
        return net_income / (number_of_people**0.5)



from policyengine_us.model_api import *


class ca_state_supplement_payment_standard(Variable):
    value_type = float
    entity = SPMUnit
    label = "California SSI state supplement payment standard"
    unit = USD
    definition_period = MONTH
    defined_for = StateCode.CA
    reference = "https://leginfo.legislature.ca.gov/faces/codes_displaySection.xhtml?lawCode=WIC&sectionNum=12200"

    adds = [
        "ca_state_supplement_dependent_amount",
        "ca_state_supplement_food_allowance",
        "ca_state_supplement_medical_care_facility_amount",
        "ca_state_supplement_out_of_home_care_facility_amount",
        "ca_state_supplement_aged_blind_disabled_amount",
    ]



from policyengine_us.model_api import *


class co_tanf_countable_gross_earned_income(Variable):
    value_type = float
    entity = SPMUnit
    label = "Colorado TANF countable gross earned income"
    unit = USD
    definition_period = YEAR

    adds = "gov.states.co.cdhs.tanf.income.earned"



from policyengine_us.model_api import *


class dc_kccatc(Variable):
    value_type = float
    entity = TaxUnit
    label = "DC keep child care affordable tax credit"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://otr.cfo.dc.gov/sites/default/files/dc/sites/otr/publication/attachments/52926_D-40_12.21.21_Final_Rev011122.pdf#page=67"
        "https://otr.cfo.dc.gov/sites/default/files/dc/sites/otr/publication/attachments/2022_D-40_Booklet_Final_blk_01_23_23_Ordc.pdf#page=59"
    )
    defined_for = StateCode.DC

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.dc.tax.income.credits
        # determine tax unit's income eligibility status
        taxinc = tax_unit("dc_taxable_income_joint", period)
        filing_status = tax_unit("filing_status", period)
        income_eligible = taxinc <= p.kccatc.income_limit[filing_status]
        # determine count of KCCATC age eligible children
        person = tax_unit.members
        is_dependent = person("is_tax_unit_dependent", period)
        age = person("age", period)
        kccatc_age_eligible = is_dependent & (age <= p.kccatc.max_age)
        kccatc_eligible_count = tax_unit.sum(kccatc_age_eligible)
        # determine count of federal CDCC age eligible children
        cdcc = parameters(period).gov.irs.credits.cdcc.eligibility
        cdcc_age_eligible = is_dependent & (age < cdcc.child_age)
        cdcc_eligible_count = tax_unit.sum(cdcc_age_eligible)
        # calculate KCCATC amount
        max_kccatc = kccatc_eligible_count * p.kccatc.max_amount
        total_care_expenses = tax_unit("tax_unit_childcare_expenses", period)
        ratio = np.zeros_like(cdcc_eligible_count)
        mask = cdcc_eligible_count > 0
        ratio[mask] = kccatc_eligible_count[mask] / cdcc_eligible_count[mask]
        kccatc_care_expenses = total_care_expenses * ratio
        kccatc = min_(kccatc_care_expenses, max_kccatc)
        # return calculated kccatc amount if income eligible
        return income_eligible * kccatc



from policyengine_us.model_api import *


class taxsim_page(Variable):
    value_type = int
    entity = TaxUnit
    label = "Age of primary taxpayer"
    unit = "year"
    documentation = "Age of primary taxpayer December 31st of the tax year (or zero). Taxpayer and spouse age variables determine eligibility for additional standard deductions, personal exemption, EITC and AMT exclusions."
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        is_head = person("is_tax_unit_head", period)
        age = person("age", period)
        return tax_unit.sum(age * is_head)



from policyengine_us.model_api import *


class ca_amt_exemption(Variable):
    value_type = float
    entity = TaxUnit
    label = "California AMT exemption amount"
    defined_for = StateCode.CA
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.ftb.ca.gov/forms/2022/2022-540-p-instructions.html"
    )

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.ca.tax.income.amt.exemption
        filing_status = tax_unit("filing_status", period)
        amti = tax_unit("ca_amti", period)
        # Instructions for Schedule P 540, line 22, Exemption Worksheet,
        # Line 1
        exemption_max_amount = p.amount[filing_status]
        # Line 2 - AMTI
        # Line 3
        lower_exemption_threshold = p.amti.threshold.lower[filing_status]
        # Line 4
        reduced_amti = max_(amti - lower_exemption_threshold, 0)
        # Line 5
        p_irs = parameters(period).gov.irs.income.amt.exemption
        reduced_amti_rate = reduced_amti * p_irs.phase_out.rate
        # Line 6
        adult_exemption = max_(exemption_max_amount - reduced_amti_rate, 0)
        # Eligible children receive an increased exemption amount
        person = tax_unit.members
        eligible_child = person("is_child_dependent", period)
        head_is_eligible_child = tax_unit.any(eligible_child)
        # Line 7
        exemption_amount_child = p_irs.child.amount
        # Line 8
        earned_income = tax_unit("head_earned", period)
        # Line 9
        total_reduced_exemption_amount = exemption_amount_child + earned_income
        # Line 10
        child_exemption = min_(total_reduced_exemption_amount, adult_exemption)
        return where(head_is_eligible_child, child_exemption, adult_exemption)



from policyengine_us.model_api import *


class ga_exemptions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Georgia Exemptions"
    defined_for = StateCode.GA
    unit = USD
    definition_period = YEAR
    reference = (
        "https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500",
        "https://advance.lexis.com/documentpage/?pdmfid=1000516&crid=2c053fd5-32c1-4cc1-86b0-36aaade9da5b&pdistocdocslideraccess=true&config=00JAA1MDBlYzczZi1lYjFlLTQxMTgtYWE3OS02YTgyOGM2NWJlMDYKAFBvZENhdGFsb2feed0oM9qoQOMCSJFX5qkd&pddocfullpath=%2Fshared%2Fdocument%2Fstatutes-legislation%2Furn%3AcontentItem%3A6348-G0H1-DYB7-W3JT-00008-00&pdcomponentid=234187&pdtocnodeidentifier=ABWAALAADAAL&ecomp=k2vckkk&prid=4862391c-e031-443f-ad52-ae86c6bb5ce2",
    )

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.ga.tax.income.exemptions
        # Dependent exemptions
        dependents = tax_unit(
            "tax_unit_dependents", period
        )  # Total the number of dependents
        dependent_exemptions = dependents * p.dependent

        # total exemptions
        if p.personal.availability:
            filing_status = tax_unit("filing_status", period)
            personal_exemptions = p.personal.amount[filing_status]
            return personal_exemptions + dependent_exemptions
        return dependent_exemptions



from policyengine_us.model_api import *


class co_income_tax_before_non_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "Colorado income tax before non-refundable credits"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.CO

    def formula(tax_unit, period, parameters):
        income = tax_unit("co_taxable_income", period)
        rate = parameters(period).gov.states.co.tax.income.rate
        return income * rate



from policyengine_us.model_api import *


class pell_grant_dependents_in_college(Variable):
    value_type = int
    entity = TaxUnit
    label = "Pell Grant dependents in college"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        college_student = person("is_full_time_college_student", period)
        is_dependent = person("is_tax_unit_dependent", period)
        return tax_unit.sum(is_dependent * college_student)



from policyengine_us.model_api import *


class cost_of_attending_college(Variable):
    value_type = float
    entity = Person
    label = "Pell Grant cost of attendance"
    definition_period = YEAR



from policyengine_us.model_api import *


class qualified_business_income(Variable):
    value_type = float
    entity = Person
    label = "Qualified business income"
    documentation = "Business income that qualifies for the qualified business income deduction."
    unit = USD
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/199A#c"
    defined_for = "business_is_qualified"

    def formula(person, period, parameters):
        p = parameters(period).gov.irs.deductions.qbi
        gross_qbi = 0
        for var in p.income_definition:
            gross_qbi += person(var, period) * person(
                var + "_would_be_qualified", period
            )
        qbi_deductions = add(person, period, p.deduction_definition)
        return max_(0, gross_qbi - qbi_deductions)



from policyengine_us.model_api import *


class ssi_blind_or_disabled_working_student_exclusion(Variable):
    value_type = float
    entity = Person
    label = "SSI blind or disabled working student earned income exclusion"
    unit = USD
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/cfr/text/20/416.1112#c_3"
    defined_for = "is_ssi_blind_or_disabled_working_student_exclusion_eligible"

    def formula(person, period, parameters):
        p = parameters(
            period
        ).gov.ssa.ssi.income.exclusions.blind_or_disabled_working_student
        monthly_earned_income = (
            person("ssi_earned_income", period) / MONTHS_IN_YEAR
        )
        max_amount = min_(p.amount, monthly_earned_income) * MONTHS_IN_YEAR
        return min_(max_amount, p.cap)



from policyengine_us.model_api import *


class meets_snap_gross_income_test(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Meets SNAP gross income test"
    documentation = "Whether this SPM unit meets the SNAP gross income test"
    definition_period = MONTH
    reference = (
        "https://www.law.cornell.edu/uscode/text/7/2017#a",
        "https://www.law.cornell.edu/uscode/text/7/2014#c",
    )

    def formula(spm_unit, period, parameters):
        limit = parameters(period).gov.usda.snap.income.limit.gross
        income = spm_unit("snap_gross_income_fpg_ratio", period)
        # Households with elderly and disabled people are exempt from the
        # gross income test.
        has_elderly_disabled = spm_unit("has_usda_elderly_disabled", period)
        return has_elderly_disabled | (income <= limit)



from policyengine_us.model_api import *


class AgeGroup(Enum):
    CHILD = "Child"
    WORKING_AGE = "Working-age"
    SENIOR = "Senior"


class age_group(Variable):
    value_type = Enum
    possible_values = AgeGroup
    default_value = AgeGroup.WORKING_AGE
    entity = Person
    label = "Age group"
    definition_period = YEAR

    def formula(person, period, parameters):
        return select(
            [
                person("is_child", period),
                person("is_wa_adult", period),
                person("is_senior", period),
            ],
            [AgeGroup.CHILD, AgeGroup.WORKING_AGE, AgeGroup.SENIOR],
        )



from policyengine_us.model_api import *


class count_days_postpartum(Variable):
    value_type = float
    entity = Person
    label = "Number of days postpartum"
    unit = "day"
    definition_period = YEAR

    def formula(person, period, parameters):
        under_60_days = person("under_60_days_postpartum", period)
        under_12_months = person("under_12_months_postpartum", period)
        return select(
            [
                under_60_days,
                under_12_months,
                True,
            ],
            [
                0,
                60,
                np.inf,
            ],
        )



from policyengine_us.model_api import *


class mi_home_heating_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Michigan home heating credit"
    defined_for = StateCode.MI
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.michigan.gov/taxes/iit/accordion/credits/table-a-2022-home-heating-credit-mi-1040cr-7-standard-allowance"
        "http://www.legislature.mi.gov/(S(keapvg1h2vndkn25rtmpyyse))/mileg.aspx?page=getObject&objectName=mcl-206-527a"
    )

    def formula(tax_unit, period, parameters):
        standard_credit = tax_unit("mi_standard_home_heating_credit", period)
        alternate_credit = tax_unit("mi_alternate_home_heating_credit", period)
        p = parameters(period).gov.states.mi.tax.income.credits.home_heating
        larger_credit = max_(standard_credit, alternate_credit)
        total_credit = larger_credit * p.credit_percentage
        rate = tax_unit("mi_home_heating_credit_eligible_rate", period)
        return total_credit * rate



from policyengine_us.model_api import *


class or_retirement_credit_household_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "Household income for the Oregon Retirement Income Tax Credit"
    unit = USD
    definition_period = YEAR
    reference = "https://casetext.com/statute/oregon-revised-statutes/title-29-revenue-and-taxation/chapter-316-personal-income-tax/additional-credits/retirement-income/section-316157-credit-for-retirement-income"
    defined_for = StateCode.OR

    adds = ["adjusted_gross_income", "tax_exempt_interest_income"]
    subtracts = ["taxable_social_security"]



from policyengine_us.model_api import *


class nd_subtractions(Variable):
    value_type = float
    entity = TaxUnit
    label = "North Dakota subtractions from federal taxable income"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2021-iit/form-nd-1-2021.pdf#page=1"  # line 16
        "https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2021-iit/individual-income-tax-booklet-2021.pdf#page=14"
        "https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/form-nd-1-2022.pdf#page=1"  # line 16
        "https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf#page=14"
    )
    defined_for = StateCode.ND

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.nd.tax.income.taxable_income.subtractions
        total_subtractions = add(tax_unit, period, p.sources)
        # Prevent negative subtractions from acting as additions
        return max_(0, total_subtractions)



from policyengine_us.model_api import *


class ct_non_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "Connecticut non-refundable credits"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.CT

    adds = "gov.states.ct.tax.income.credits.non_refundable"



from policyengine_us.model_api import *


class id_retirement_benefits_deduction_eligible_person(Variable):
    value_type = bool
    entity = Person
    label = "Eligible person for the Idaho retirement benefits deduction"
    reference = "https://legislature.idaho.gov/statutesrules/idstat/title63/t63ch30/sect63-3022a/"
    definition_period = YEAR
    defined_for = StateCode.ID

    def formula(person, period, parameters):
        p = parameters(
            period
        ).gov.states.id.tax.income.deductions.retirement_benefits

        age_threshold = where(
            person("is_disabled", period),
            p.age_eligibility.disabled,
            p.age_eligibility.main,
        )
        age_eligible = person("age", period) >= age_threshold

        head_or_spouse = person("is_tax_unit_head_or_spouse", period)

        return head_or_spouse & age_eligible



from policyengine_us.model_api import *


class is_fully_disabled_service_connected_veteran(Variable):
    value_type = bool
    entity = Person
    label = "Is a fully disabled veteran who became so as a result of an injury during service"
    definition_period = YEAR



from policyengine_us.model_api import *


class is_ccdf_eligible(Variable):
    value_type = bool
    entity = Person
    definition_period = YEAR
    label = "Eligibility for CCDF"

    def formula(person, period, parameters):
        asset_eligible = person.spm_unit("is_ccdf_asset_eligible", period)
        age_eligible = person("is_ccdf_age_eligible", period)
        income_eligible = person.spm_unit("is_ccdf_income_eligible", period)
        reason_for_care_eligible = person(
            "is_ccdf_reason_for_care_eligible", period
        )

        return (
            asset_eligible
            & age_eligible
            & income_eligible
            & reason_for_care_eligible
        )



from policyengine_us.model_api import *


class va_low_income_tax_credit_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Eligible for the Virginia Low Income Tax Credit"
    definition_period = YEAR
    defined_for = StateCode.VA

    def formula(tax_unit, period, parameters):
        # Criteria 1: Not qualified to claim EITC if anyone in the tax unit claimed any of the following:
        p = parameters(
            period
        ).gov.states.va.tax.income.credits.eitc.low_income_tax
        program_eligible = add(tax_unit, period, p.ineligible_programs) == 0

        # Criteria 2: Not qualified to claim EITC if the filier is claimed as a dependent on another taxpayer's return

        head_is_not_dependent_elsewhere = ~tax_unit(
            "head_is_dependent_elsewhere", period
        )

        return program_eligible & head_is_not_dependent_elsewhere



from policyengine_us.model_api import *


class pa_tanf_age_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Pennsylvania TANF age eligibility"
    definition_period = YEAR
    defined_for = StateCode.PA

    def formula(spm_unit, period, parameters):
        person = spm_unit.members
        p = parameters(period).gov.states.pa.dhs.tanf.cash_assistance
        age = person("age", period)
        # Get full time students
        student_eligible = person("is_full_time_student", period) & (
            age == p.age_limit
        )

        # Get age
        is_eligible_age = age < p.age_limit
        return spm_unit.any(is_eligible_age | student_eligible)



from policyengine_us.model_api import *


class dwks19(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "IRS Form 1040 Schedule D worksheet (part 6 of 6)"
    unit = USD

    def formula(tax_unit, period, parameters):
        dwks14 = tax_unit("dwks14", period)
        capital_gains = parameters(period).gov.irs.capital_gains.brackets
        filing_status = tax_unit("filing_status", period)
        dwks01 = tax_unit("taxable_income", period)
        dwks16 = min_(capital_gains.thresholds["1"][filing_status], dwks01)
        dwks17 = min_(dwks14, dwks16)
        dwks10 = tax_unit("dwks10", period)
        dwks18 = max_(0, dwks01 - dwks10)
        return max_(dwks17, dwks18) * tax_unit("has_qdiv_or_ltcg", period)



from policyengine_us.model_api import *


class oh_exemption_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Ohio Exemption Credit"
    unit = USD
    definition_period = YEAR
    reference = "https://law.justia.com/codes/ohio/2022/title-57/chapter-5747/section-5747-022/"
    defined_for = StateCode.OH

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.oh.tax.income.credits.exemption

        agi = tax_unit("oh_agi", period)
        personal_exemptions = tax_unit("oh_personal_exemptions", period)
        # Per tax form, amount can be negative
        modified_agi = agi - personal_exemptions
        amount_per_exemption = p.amount.calc(modified_agi, right=True)

        exemptions = tax_unit("exemptions_count", period)

        return amount_per_exemption * exemptions



from policyengine_us.model_api import *


class wi_childcare_expense_subtraction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Wisconsin childcare expense subtraction from federal AGI"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.revenue.wi.gov/TaxForms2021/2021-ScheduleSB.pdf"
        "https://www.revenue.wi.gov/TaxForms2021/2021-ScheduleSB-inst.pdf#page=7"
    )
    defined_for = StateCode.WI

    def formula(tax_unit, period, parameters):
        subtraction = parameters(period).gov.states.wi.tax.income.subtractions
        uncapped_expenses = tax_unit("tax_unit_childcare_expenses", period)
        eligible_dependents = tax_unit("count_cdcc_eligible", period)
        count_eligible = min_(
            eligible_dependents, subtraction.childcare_expense.max_dependents
        )
        capped_expenses = min_(
            uncapped_expenses,
            count_eligible * subtraction.childcare_expense.max_amount,
        )
        return min_(
            capped_expenses, tax_unit("min_head_spouse_earned", period)
        )



from policyengine_us.model_api import *


class is_chip_eligible_child(Variable):
    value_type = bool
    entity = Person
    label = "Child eligible for CHIP"
    documentation = "Determines if a child is eligible for the Children's Health Insurance Program"
    definition_period = YEAR
    reference = (
        "https://www.ssa.gov/OP_Home/ssact/title21/2110.htm",
        "https://www.medicaid.gov/medicaid/national-medicaid-chip-program-information/medicaid-childrens-health-insurance-program-basic-health-program-eligibility-levels",
    )

    def formula(person, period, parameters):
        # Get state code
        state_code = person.household("state_code", period)

        # Check age eligibility
        age = person("age", period)
        p = parameters(period).gov.hhs.chip.child

        is_age_eligible = age < p.max_age

        # Check if state offers CHIP program for children
        income_limit = p.income_limit[state_code]
        state_has_chip = income_limit > 0

        # Check immigration status eligibility
        istatus = person("immigration_status", period)
        undocumented = istatus == istatus.possible_values.UNDOCUMENTED
        immigration_eligible = ~undocumented

        # Check income eligibility
        # CHIP is for children who make too much for Medicaid but below CHIP limits
        # First, check if not eligible for Medicaid
        medicaid_eligible = person("is_medicaid_eligible", period)

        # Check if family income is below CHIP threshold
        # Use medicaid_income_level as the income measure
        income_ratio = person("medicaid_income_level", period)
        income_eligible = income_ratio <= income_limit

        return (
            is_age_eligible
            & state_has_chip
            & immigration_eligible
            & ~medicaid_eligible
            & income_eligible
        )



from policyengine_us.model_api import *


class taxsim_v11(Variable):
    value_type = float
    entity = TaxUnit
    label = "UI in AGI"
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return tax_unit("tax_unit_taxable_unemployment_compensation", period)



from policyengine_us.model_api import *


class dc_gac_countable_unearned_income_person(Variable):
    value_type = float
    entity = Person
    label = (
        "DC General Assistance for Children (GAC) unearned income per person"
    )
    unit = USD
    definition_period = MONTH
    reference = (
        "https://code.dccouncil.gov/us/dc/council/code/sections/4-205.05a#(e)"
    )
    defined_for = "dc_gac_eligible_child"

    adds = ["dc_tanf_gross_unearned_income"]



from policyengine_us.model_api import *


class ri_cdcc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Rhode Island child and dependent care credit"
    defined_for = StateCode.RI
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        # Rhode Island matches the federal credit taken
        fed_cdcc = tax_unit("cdcc", period)
        rate = parameters(period).gov.states.ri.tax.income.credits.cdcc.rate
        return fed_cdcc * rate



from policyengine_us.model_api import *


class ma_part_b_taxable_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "MA Part B taxable income"
    unit = USD
    definition_period = YEAR
    reference = "https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions"
    defined_for = StateCode.MA

    def formula(tax_unit, period, parameters):
        taxinc_before_exemption = tax_unit(
            "ma_part_b_taxable_income_before_exemption", period
        )
        total_exemption = tax_unit(
            "ma_part_b_taxable_income_exemption", period
        )
        return max_(0, taxinc_before_exemption - total_exemption)



from policyengine_us.model_api import *


class is_head_start_eligible(Variable):
    value_type = bool
    entity = Person
    label = "Eligible person for the Head Start program"
    definition_period = YEAR
    reference = (
        "https://eclkc.ohs.acf.hhs.gov/policy/45-cfr-chap-xiii/1302-12-determining-verifying-documenting-eligibility"
        "https://www.hhs.gov/answers/programs-for-families-and-children/how-can-i-get-my-child-into-head-start/index.html"
    )

    def formula(person, period, parameters):
        p = parameters(period).gov.hhs.head_start

        age = person("age", period)
        is_age_eligible = p.age_range.calc(age)
        is_program_eligible = person(
            "is_head_start_categorically_eligible", period
        )
        is_income_eligible = person("is_head_start_income_eligible", period)

        return is_age_eligible & (is_income_eligible | is_program_eligible)



from policyengine_us.model_api import *


class Race(Enum):
    WHITE = "White, non-Hispanic"
    BLACK = "Black, non-Hispanic"
    HISPANIC = "Hispanic"
    OTHER = "Other"


class race(Variable):
    value_type = Enum
    possible_values = Race
    default_value = Race.WHITE
    entity = Person
    label = "race"
    documentation = "The broadest racial category (White only, Black only, Hispanic, Other)"
    definition_period = YEAR

    def formula(person, period, parameters):
        cps_race = person("cps_race", period)
        hispanic = person("is_hispanic", period)
        return select(
            [
                hispanic,
                (cps_race == 1) & ~hispanic,
                (cps_race == 2) & ~hispanic,
                True,
            ],
            [
                Race.HISPANIC,
                Race.WHITE,
                Race.BLACK,
                Race.OTHER,
            ],
        )



from policyengine_us.model_api import *


class ca_tanf_immigration_status_eligible_person(Variable):
    value_type = bool
    entity = Person
    label = "California TANF immigration status Eligibility"
    definition_period = YEAR
    defined_for = StateCode.CA
    reference = "http://epolicy.dpss.lacounty.gov/epolicy/epolicy/server/general/projects_responsive/ePolicyMaster/index.htm?&area=general&type=responsivehelp&ctxid=&project=ePolicyMaster#t=mergedProjects%2FChild%20Care%2FChild_Care%2F1210_Overview%2F1210_Overview.htm%23Backgroundbc-3&rhtocid=_3_3_0_2"

    def formula(person, period, parameters):
        immigration_status = person("immigration_status", period)
        p = parameters(period).gov.states.ca.cdss.tanf.cash.immigration_status

        immigration_status_str = immigration_status.decode_to_str()

        return np.isin(
            immigration_status_str,
            p.eligible_statuses,
        )



from policyengine_us.model_api import *


class is_pregnant_for_medicaid_nfc(Variable):
    value_type = bool
    entity = Person
    label = "Medicaid pregnant non-financial criteria"
    definition_period = YEAR

    def formula(person, period, parameters):
        ma = parameters(
            period
        ).gov.hhs.medicaid.eligibility.categories.pregnant
        is_pregnant = person("is_pregnant", period)
        days_postpartum = person("count_days_postpartum", period)
        state = person.household("state_code_str", period)
        max_postpartum_days = ma.postpartum_coverage[state]
        return is_pregnant | (days_postpartum < max_postpartum_days)



from policyengine_us.model_api import *


class adjusted_gross_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "Adjusted gross income"
    unit = USD
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/62"

    def formula(tax_unit, period, parameters):
        gross_income = add(tax_unit, period, ["irs_gross_income"])
        above_the_line_deductions = tax_unit(
            "above_the_line_deductions", period
        )
        agi = gross_income - above_the_line_deductions
        if parameters(period).gov.contrib.ubi_center.basic_income.taxable:
            agi += add(tax_unit, period, ["basic_income"])
        return agi



from policyengine_us.model_api import *


class ca_calworks_child_care_payment_standard(Variable):
    value_type = float
    entity = Person
    label = "California CalWORKs Child Care Payment Standard"
    unit = USD
    definition_period = MONTH
    defined_for = "ca_calworks_child_care_child_age_eligible"
    reference = "http://epolicy.dpss.lacounty.gov/epolicy/epolicy/server/general/projects_responsive/ePolicyMaster/index.htm?&area=general&type=responsivehelp&ctxid=&project=ePolicyMaster#t=mergedProjects%2FChild%20Care%2FChild_Care%2F1210_8_Regional_Market_Rate_Ceilings%2F1210_8_Regional_Market_Rate_Ceilings.htm%23Contactbc-13&rhtocid=_3_3_8_12"

    def formula(person, period, parameters):
        p = parameters(period).gov.states.ca.cdss.tanf.child_care.rate_ceilings
        # Payment standard is the maximum payment depending on:
        # - Period (hourly/daily/weekly/monthly)
        # - Care facility type
        # - Full-time/part-time
        provider = person("ca_calworks_child_care_provider_category", period)
        time = person("ca_calworks_child_care_time_category", period)
        is_full_time = person("ca_calworks_child_care_full_time", period)
        age = person("age", period.this_year)

        age_group = select(
            [age < p.age_threshold.lower, age > p.age_threshold.higher],
            ["younger", "older"],
            default="middle",
        )
        return p.standard[age_group][provider][time][is_full_time]



from policyengine_us.model_api import *


class recapture_of_investment_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Recapture of Investment Credit"
    unit = USD
    definition_period = YEAR



from policyengine_us.model_api import *


class ca_standard_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "California standard deduction"
    unit = USD
    documentation = "https://www.ftb.ca.gov/forms/2021/2021-540.pdf"
    definition_period = YEAR
    defined_for = StateCode.CA

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.ca.tax.income.deductions.standard
        filing_status = tax_unit("filing_status", period)
        return p.amount[filing_status]



from policyengine_us.model_api import *


class is_ccdf_income_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    definition_period = YEAR
    label = "Income eligibility for CCDF"

    def formula(spm_unit, period, parameters):
        income_to_smi_ratio = spm_unit("ccdf_income_to_smi_ratio", period)
        p_ratio_limit = parameters(period).gov.hhs.ccdf.income_limit_smi
        return income_to_smi_ratio <= p_ratio_limit



from policyengine_us.model_api import *


class pell_grant_head_assets(Variable):
    value_type = float
    entity = TaxUnit
    unit = USD
    label = "Pell Grant head assets"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        is_head = person("is_tax_unit_head", period)
        is_spouse = person("is_tax_unit_spouse", period)
        assets = person("pell_grant_countable_assets", period)
        is_parent = is_head | is_spouse
        return tax_unit.sum(is_parent * assets)



from policyengine_us.model_api import *


class ILPersonalExemptionEligibilityStatus(Enum):
    BOTH_ELIGIBLE = 1
    PARTIALLY_ELIGIBLE = 2
    NOT_ELIGIBLE = 3


class il_personal_exemption_eligibility_status(Variable):
    value_type = Enum
    possible_values = ILPersonalExemptionEligibilityStatus
    default_value = ILPersonalExemptionEligibilityStatus.NOT_ELIGIBLE
    entity = TaxUnit
    label = (
        "Whether The Tax Unit Is Eligible For The Illinois Personal Exemption"
    )
    definition_period = YEAR
    defined_for = StateCode.IL

    def formula(tax_unit, period, parameters):
        personal_eligibility_amount = parameters(
            period
        ).gov.states.il.tax.income.exemption.personal

        # First, determine whether the tax unit is filing jointly or not
        filing_status = tax_unit("filing_status", period)
        joint = filing_status == filing_status.possible_values.JOINT

        tax_unit_personal_eligibility_amount = (
            personal_eligibility_amount * where(joint, 2, 1)
        )

        # Then, determine whether either the head or the spouse of the
        # tax unit is claimable as a dependent in another unit
        claimable_count = add(
            tax_unit,
            period,
            ["head_is_dependent_elsewhere", "spouse_is_dependent_elsewhere"],
        )
        il_base_income = tax_unit("il_base_income", period)

        # Criteria for complete ineligibility
        ineligible = (
            (
                ~joint
                & (claimable_count > 0)
                & (il_base_income > tax_unit_personal_eligibility_amount)
            )
            | (
                joint
                & (claimable_count > 1)
                & (il_base_income > tax_unit_personal_eligibility_amount)
            )
            | (
                joint
                & (claimable_count == 2)
                & (il_base_income > tax_unit_personal_eligibility_amount * 2)
            )
        )

        # Criteria for partial ineligibility
        partner_ineligible = (
            (
                joint
                & (claimable_count == 1)
                & (il_base_income > tax_unit_personal_eligibility_amount)
            )
            | (~joint & (claimable_count == 0))
            | (
                ~joint
                & (claimable_count == 1)
                & (il_base_income < tax_unit_personal_eligibility_amount)
            )
        )

        # Based on the criteria, return the eligibility status
        return select(
            [ineligible, partner_ineligible],
            [
                ILPersonalExemptionEligibilityStatus.NOT_ELIGIBLE,
                ILPersonalExemptionEligibilityStatus.PARTIALLY_ELIGIBLE,
            ],
            ILPersonalExemptionEligibilityStatus.BOTH_ELIGIBLE,
        )



from policyengine_us.model_api import *


class adjusted_net_capital_gain(Variable):
    value_type = float
    entity = TaxUnit
    label = "Adjusted net capital gain"
    unit = USD
    documentation = "The excess of net long-term capital gain over net short-term capital loss."
    definition_period = YEAR
    reference = dict(
        title="26 U.S. Code § 1(h)(3)",
        href="https://www.law.cornell.edu/uscode/text/26/1#h_3",
    )

    def formula(tax_unit, period, parameters):
        net_capital_gain = tax_unit("net_capital_gain", period)
        # The law actually uses the original definition of 'net capital gain' which does not include
        # qualified dividend income, but separately adds qualified dividends here. The definition of
        # 'net capital gain' in the variable 'net_capital_gain' actually has some very specific exclusion
        # criteria for particular types of dividends and companies, so it's not an *exact* fit to the
        # definition here, but it's a good enough approximation. See 26 U.S. Code § 1(h)(11)(B) for the
        # definition of 'net capital gain' for the above variable, and 26 U.S. Code § 1(h)(3) for the definition
        # of adjusted net capital gain (this variable).
        qualified_dividend_income = add(
            tax_unit, period, ["qualified_dividend_income"]
        )
        unrecaptured_s_1250_gain = tax_unit(
            "unrecaptured_section_1250_gain", period
        )
        cg_28_pct_rate_gain = tax_unit(
            "capital_gains_28_percent_rate_gain", period
        )
        net_gains_less_dividends = max_(
            0,
            net_capital_gain - qualified_dividend_income,
        )
        reduced_capital_gains = max_(
            net_gains_less_dividends
            - (unrecaptured_s_1250_gain + cg_28_pct_rate_gain),
            0,
        )
        return reduced_capital_gains + qualified_dividend_income



from policyengine_us.model_api import *


class interest_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Interest deduction"
    unit = USD
    documentation = "Interest expenses deducted from taxable income."
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/163"

    adds = ["deductible_interest_expense"]



from policyengine_us.model_api import *


class ma_eaedc(Variable):
    value_type = float
    entity = SPMUnit
    label = "Massachusetts EAEDC"
    unit = USD
    definition_period = MONTH
    defined_for = "ma_eaedc_eligible"
    reference = "https://www.law.cornell.edu/regulations/massachusetts/department-106-CMR/title-106-CMR-701.000"

    def formula(spm_unit, period, parameters):
        tafdc_exceeds_eaedc = spm_unit("ma_tafdc_exceeds_eaedc", period)
        eaedc_benefit_amount = spm_unit("ma_eaedc_if_claimed", period)
        return ~tafdc_exceeds_eaedc * eaedc_benefit_amount



from policyengine_us.model_api import *


class ma_tafdc_dependent_criteria_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Eligible based on the dependent criteria for the Massachusetts Temporary Assistance for Families with Dependent Children (TAFDC)"
    definition_period = YEAR
    reference = (
        "https://www.law.cornell.edu/regulations/massachusetts/106-CMR-703-210"
    )
    defined_for = StateCode.MA

    def formula(spm_unit, period, parameters):
        dependent_child_present = (
            add(spm_unit, period, ["ma_tafdc_eligible_dependent"]) > 0
        )
        dependent_child_pregnancy_eligible = spm_unit.any(
            spm_unit.members("ma_tafdc_pregnancy_eligible", period)
        )  # adult pregnancy included
        return dependent_child_present | dependent_child_pregnancy_eligible



from policyengine_us.model_api import *


class is_ca_medicaid_immigration_status_eligible(Variable):
    value_type = bool
    entity = Person
    label = "Eligible for California state-funded Medicaid regardless of immigration status"
    definition_period = YEAR
    reference = [
        "https://california.public.law/codes/welfare_and_institutions_code_section_14007.8",
    ]
    defined_for = StateCode.CA

    def formula(person, period, parameters):
        p = parameters(period).gov.states.ca.chhs
        age = person("age", period)
        pregnant = person("is_pregnant", period)
        return p.eligible_regardless_of_immigration_status.calc(age) | pregnant



from policyengine_us.model_api import *


class ca_riv_general_relief_countable_income_person(Variable):
    value_type = float
    entity = Person
    unit = USD
    label = "Person's countable income before proration"
    definition_period = MONTH

    adds = [
        "ca_riv_general_relief_net_earned_income",
        "ca_riv_general_relief_unearned_income",
    ]



from policyengine_us.model_api import *


class ks_itemized_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Kansas itemized deductions"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.ksrevenue.gov/pdf/ip21.pdf"
        "https://www.ksrevenue.gov/pdf/ip22.pdf"
    )
    defined_for = StateCode.KS

    def formula(tax_unit, period, parameters):
        # 2021 Form K-40 instructions say this:
        #   LINE 4 (Standard deduction or itemized deductions):
        #   If you did not itemize your deductions on your federal return,
        #   you may choose to itemize your deductions or claim the
        #   standard deduction on your Kansas return whichever is to your
        #   advantage.  If you itemized on your federal return, you may
        #   either itemize or take the standard deduction on your Kansas
        #   return, whichever is to your advantage.
        # compute itemized deduction maximum
        itm_deds_less_salt = tax_unit("itemized_deductions_less_salt", period)
        uncapped_property_taxes = add(tax_unit, period, ["real_estate_taxes"])
        return itm_deds_less_salt + uncapped_property_taxes



from policyengine_us.model_api import *


class ca_la_infant_supplement_eligible_person(Variable):
    value_type = bool
    entity = Person
    definition_period = MONTH
    label = "Eligible for the Los Angeles County infant supplement"
    defined_for = "in_la"

    def formula(person, period, parameters):
        return person("ca_foster_care_minor_dependent", period)



from policyengine_us.model_api import *


class is_wa_adult(Variable):
    value_type = bool
    entity = Person
    label = "Is a working-age adult"
    definition_period = YEAR

    def formula(person, period, parameters):
        return ~person("is_child", period) & ~person("is_senior", period)



from policyengine_us.model_api import *


class wi_homestead_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Wisconsin homestead credit"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.revenue.wi.gov/TaxForms2021/2021-Form1f.pdf#page=3"
        "https://www.revenue.wi.gov/TaxForms2021/2021-Form1-Inst.pdf#page=27"
        "https://www.revenue.wi.gov/TaxForms2022/2022-Form1f.pdf#page=3"
        "https://www.revenue.wi.gov/TaxForms2022/2022-Form1-Inst.pdf#page=28"
        "https://docs.legis.wisconsin.gov/misc/lfb/informational_papers/january_2023/0002_individual_income_tax_informational_paper_2.pdf"
    )
    defined_for = "wi_homestead_eligible"

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.wi.tax.income.credits
        uncapped_ptax = tax_unit("wi_homestead_property_tax", period)
        capped_ptax = min_(p.homestead.property_tax.max, uncapped_ptax)
        hincome = tax_unit("wi_homestead_income", period)
        phase_out_start = p.homestead.phase_out.start
        phase_out_rate = p.homestead.phase_out.rate
        phase_out = max_(0, hincome - phase_out_start) * phase_out_rate
        return max_(0, capped_ptax - phase_out) * p.homestead.rate



from policyengine_us.model_api import *


class id_pbf(Variable):
    value_type = float
    entity = TaxUnit
    label = "Idaho permanent building tax"
    definition_period = YEAR
    reference = "https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_11-15-2021.pdf#page=10"
    defined_for = "id_pbf_liable"

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.id.tax.income.other_taxes.pbf
        return p.amount



from policyengine_us.model_api import *


class ny_tanf_countable_earned_income(Variable):
    value_type = float
    entity = SPMUnit
    label = "New York TANF countable earned income"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.NY

    def formula(spm_unit, period, parameters):
        # Get gross earned income.
        gross_earned_income = spm_unit("ny_tanf_gross_earned_income", period)
        # Multiply by 100% minus the EID.
        p = parameters(
            period
        ).gov.states.ny.otda.tanf.income.earned_income_deduction
        # https://otda.ny.gov/policy/gis/2022/22DC085.pdf Section 2
        # Effective October 1, 2022, the EID will be applied prior to the work expense disregard.
        return max_(
            gross_earned_income * (1 - p.percent) - p.flat * MONTHS_IN_YEAR,
            0,
        )



from policyengine_us.model_api import *


class me_non_refundable_child_care_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Maine non-refundable child care credit"
    unit = USD
    documentation = (
        "The portion of the Maine Child Care Credit that is non-refundable."
    )
    reference = [
        "https://www.mainelegislature.org/legis/statutes/36/title36sec5218.html",
        "https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_sched_a_ff.pdf#page=2",
    ]
    definition_period = YEAR
    defined_for = StateCode.ME

    def formula(tax_unit, period, parameters):
        child_care_credit = tax_unit("me_child_care_credit", period)
        refundable_child_care_credit = tax_unit(
            "me_refundable_child_care_credit", period
        )

        return max_(child_care_credit - refundable_child_care_credit, 0)



from policyengine_us.model_api import *


class hi_tax_credit_for_low_income_household_renters(Variable):
    value_type = float
    entity = TaxUnit
    label = "Hawaii low income household renters tax credit"
    unit = USD
    definition_period = YEAR
    defined_for = "hi_tax_credit_for_low_income_household_renters_eligible"

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.hi.tax.income.credits.lihrtc

        # Aged extra exemptions
        aged_head = (
            tax_unit("age_head", period) >= p.aged_age_threshold
        ).astype(int)
        aged_spouse = (
            tax_unit("age_spouse", period) >= p.aged_age_threshold
        ).astype(int)

        aged_exemptions = aged_head + aged_spouse
        exemptions = tax_unit("exemptions_count", period)

        total_exemptions = exemptions + aged_exemptions
        return p.amount * total_exemptions



from policyengine_us.model_api import *


class ssi_amount_if_eligible(Variable):
    value_type = float
    entity = Person
    label = "SSI amount if eligible"
    unit = USD
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/42/1382#b"

    def formula(person, period, parameters):
        p = parameters(period).gov.ssa.ssi.amount
        is_dependent = person("is_tax_unit_dependent", period)
        head_or_spouse_amount = where(
            person("ssi_claim_is_joint", period),
            p.couple / 2,
            p.individual,
        )
        # Adults amount is based on whether it is a joint claim
        # Dependents always use individual amount.
        ssi_per_month = where(
            is_dependent, p.individual, head_or_spouse_amount
        )
        return ssi_per_month * MONTHS_IN_YEAR



from policyengine_us.model_api import *


class ssi_ineligible_child_allocation(Variable):
    value_type = float
    entity = Person
    label = "SSI ineligible child allocation"
    unit = USD
    documentation = "The amount of income that SSI deems ought to be spent on this child, and therefore is not deemed to SSI claimants."
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/cfr/text/20/416.1163"

    def formula(person, period, parameters):
        income = add(
            person, period, ["ssi_earned_income", "ssi_unearned_income"]
        )
        ssi = parameters(period).gov.ssa.ssi.amount
        allocation = (
            person("is_ssi_ineligible_child", period)
            * (ssi.couple - ssi.individual)
            * MONTHS_IN_YEAR
        )
        return max_(0, allocation - income)



from policyengine_us.model_api import *


class tx_fpp_income_limit(Variable):
    value_type = float
    entity = SPMUnit
    label = "Texas Family Planning Program income limit"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.healthytexaswomen.org/healthcare-programs/family-planning-program/fpp-who-can-apply",
        "https://www.hhs.texas.gov/sites/default/files/documents/texas-womens-health-programs-report-2024.pdf",
    )
    defined_for = StateCode.TX

    def formula(spm_unit, period, parameters):
        # Get the Federal Poverty Guideline for this SPM unit
        fpg = spm_unit("spm_unit_fpg", period)

        # Get FPG percentage (250% = 2.5)
        p = parameters(period).gov.states.tx.fpp

        # Return annual income limit (FPG * percentage)
        return fpg * p.fpg_percentage



from policyengine_us.model_api import *


class dc_tanf_work_requirement_exempt(Variable):
    value_type = bool
    entity = Person
    definition_period = MONTH
    label = "Exempted from working requirement for DC Temporary Assistance for Needy Families (TANF)"
    reference = (
        "https://code.dccouncil.gov/us/dc/council/code/sections/4-205.19g"
    )
    defined_for = StateCode.DC

    def formula(person, period, parameters):
        p = parameters(
            period
        ).gov.states.dc.dhs.tanf.work_requirement.work_exempted
        age = person("monthly_age", period)
        # (1) Eligible child who are not the head of an assistance unit can exempt
        eligible_child = person("dc_pap_eligible_child", period)
        # (2) Single parent works more than 30 hours per week (modeled in work requirements)
        # (3) Two-parent work more than 35 hours per week (modeled in work requirements)
        # (4) Single parent can exempt with infant less than 1 years old can exempt
        filing_status = person.tax_unit("filing_status", period)
        joint = filing_status == filing_status.possible_values.JOINT
        has_infant = person.spm_unit.any(age < p.infant_age_threshold)
        is_head = person("is_tax_unit_head", period)
        single_parent_exempt = ~joint & has_infant & is_head
        # (5) Person older than 60 can exempt
        elderly = age > p.elderly_age_threshold
        # (6) Individuals who are enrolled in post-secondary educational institutions
        is_college_student = person("is_full_time_college_student", period)

        return (
            eligible_child
            | single_parent_exempt
            | elderly
            | is_college_student
        )



from policyengine_us.model_api import *


class dc_standard_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "DC standard deduction"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.DC
    adds = ["basic_standard_deduction", "additional_standard_deduction"]



from policyengine_us.model_api import *


class household_health_benefits(Variable):
    value_type = float
    entity = Household
    label = "Household health benefits"
    unit = USD
    definition_period = YEAR

    def formula(household, period, parameters):
        p = parameters(period)
        if p.gov.simulation.include_health_benefits_in_net_income:
            return add(
                household, period, p.gov.household.household_health_benefits
            )
        else:
            return 0



from policyengine_us.model_api import *


class oh_other_add_backs(Variable):
    value_type = float
    entity = Person
    label = "Ohio other add backs"
    definition_period = YEAR
    defined_for = StateCode.OH
    documentation = ""
    reference = ""



from policyengine_us.model_api import *


class ct_child_tax_rebate(Variable):
    value_type = float
    entity = TaxUnit
    label = "Connecticut child tax rebate"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.CT

    def formula(tax_unit, period, parameters):
        income = tax_unit("ct_agi", period)
        filing_status = tax_unit("filing_status", period)
        p = parameters(period).gov.states.ct.tax.income.rebate

        reduction_start = p.reduction.start[filing_status]
        increment = p.reduction.increment
        reduction_per_increment = p.reduction.rate * increment

        excess = max_(income - reduction_start, 0)
        increments = np.ceil(excess / increment)
        total_reduction_amount = increments * reduction_per_increment

        count_dependents = tax_unit("tax_unit_count_dependents", period)
        capped_children = min_(count_dependents, p.child_cap)
        total_rebate = capped_children * p.amount

        return max_(total_rebate - total_reduction_amount, 0)



from policyengine_us.model_api import *


class ct_personal_credit_rate(Variable):
    value_type = float
    entity = TaxUnit
    unit = "\1"
    label = "Connecticut personal credit rate"
    definition_period = YEAR
    defined_for = StateCode.CT

    def formula(tax_unit, period, parameters):
        agi = tax_unit("ct_agi", period)
        filing_status = tax_unit("filing_status", period)
        status = filing_status.possible_values
        p = parameters(period).gov.states.ct.tax.income.agi

        return select(
            [
                filing_status == status.SINGLE,
                filing_status == status.JOINT,
                filing_status == status.SEPARATE,
                filing_status == status.SURVIVING_SPOUSE,
                filing_status == status.HEAD_OF_HOUSEHOLD,
            ],
            [
                p.single.calc(agi, right=True),
                p.joint.calc(agi, right=True),
                p.separate.calc(agi, right=True),
                p.surviving_spouse.calc(agi, right=True),
                p.head_of_household.calc(agi, right=True),
            ],
        )



from policyengine_us.model_api import *


class spm_unit_is_married(Variable):
    value_type = bool
    entity = SPMUnit
    label = "SPM unit is married"
    documentation = "Whether the adults in this SPM unit are married."
    definition_period = YEAR

    def formula(spm_unit, period, parameters):
        person = spm_unit.members
        return spm_unit.any(person("is_tax_unit_spouse", period))



from policyengine_us.model_api import *


class va_age_deduction_agi(Variable):
    value_type = float
    entity = TaxUnit
    label = "Virginia adjusted gross income for the age deduction"
    unit = USD
    definition_period = YEAR
    reference = [
        "https://law.lis.virginia.gov/vacodefull/title58.1/chapter3/article2/",  # § 58.1-322.03.(5)(b)
        "https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf#page=17",
    ]
    defined_for = StateCode.VA

    adds = ["adjusted_gross_income"]
    subtracts = ["tax_unit_taxable_social_security"]



from policyengine_us.model_api import *


class la_itemized_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Louisiana itemized deductions"
    unit = USD
    definition_period = YEAR
    reference = [
        "https://revenue.louisiana.gov/TaxForms/IT540iWEB(2022)D1.pdf#page=2",  # 2022 line 8B-line 8C
        "https://revenue.louisiana.gov/TaxForms/IT540i(2021)%20Instructions.pdf#page=3",  # 2021 line 8A-line 8C
        "https://www.legis.la.gov/Legis/Law.aspx?d=101760",  # (3)
    ]
    defined_for = StateCode.LA

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.la.tax.income.deductions.itemized
        federal_standard_deduction = tax_unit("standard_deduction", period)
        # Louisiana limits the itemized deductions to the amount of either the federal medical
        # expense deduction or the total federal itemized deductions
        # less the federal standard deduction
        relevant_itemized_deductions = add(
            tax_unit, period, p.relevant_federal_deductions
        )
        reduced_relevant_itemized_deductions = max_(
            relevant_itemized_deductions - federal_standard_deduction, 0
        )
        return reduced_relevant_itemized_deductions * p.excess_fraction



from policyengine_us.model_api import *


class md_income_tax_before_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "MD income tax after non-refundable credits"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MD

    def formula(tax_unit, period, parameters):
        tax_before_credits = tax_unit("md_income_tax_before_credits", period)
        non_refundable_credits = tax_unit("md_non_refundable_credits", period)
        return max_(0, tax_before_credits - non_refundable_credits)



from policyengine_us.model_api import *


class nj_potential_property_tax_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "New Jersey potential property tax deduction"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://law.justia.com/codes/new-jersey/2022/title-54a/section-54a-3a-17/"
        "https://www.state.nj.us/treasury/taxation/pdf/other_forms/tgi-ee/2021/1040i.pdf#page=25"
    )
    defined_for = "nj_property_tax_deduction_eligible"

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.nj.tax.income.deductions.property_tax

        # property_tax amount calculation follows NJ-1040 form Worksheet G
        rent = add(tax_unit, period, ["rent"])
        ptax = add(tax_unit, period, ["real_estate_taxes"])
        property_tax = ptax + rent * p.qualifying_rent_fraction

        # if filing separate but maintain same home, halve property_tax amount
        filing_status = tax_unit("filing_status", period)
        separate = filing_status == filing_status.possible_values.SEPARATE
        cohabitating = tax_unit("cohabitating_spouses", period)
        property_tax = property_tax / (1 + separate * cohabitating)

        # limit property_tax amount
        limit = p.limit / (1 + separate * cohabitating)
        return min_(property_tax, limit)



from policyengine_us.model_api import *


class ma_tafdc_exceeds_eaedc(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Whether the TAFDC value exceeds the EAEDC value"
    definition_period = MONTH
    reference = "https://www.mass.gov/how-to/transitional-aid-to-families-with-dependent-children-tafdc"
    defined_for = StateCode.MA

    def formula(spm_unit, period, parameters):
        tafdc_value = spm_unit("ma_tafdc_if_claimed", period)
        eaedc_value = spm_unit("ma_eaedc_if_claimed", period)
        return tafdc_value > eaedc_value



from policyengine_us.model_api import *


class va_agi_share(Variable):
    value_type = float
    entity = Person
    label = "Virginia share of state adjusted gross income of each person"
    unit = USD
    definition_period = YEAR
    reference = "https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2021-760-instructions.pdf#page=20"
    defined_for = StateCode.VA

    def formula(person, period, parameters):
        agi = person("va_agi_person", period)
        total_agi = person.tax_unit.sum(agi)
        # avoid divide-by-zero warnings
        fraction = np.zeros_like(total_agi)
        mask = total_agi != 0
        fraction[mask] = agi[mask] / total_agi[mask]
        # if no net income, then assign entirely to head.
        return where(
            total_agi == 0,
            person("is_tax_unit_head", period),
            fraction,
        )



from policyengine_us.model_api import *


class general_business_credit(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    label = "General business credit"
    documentation = "General business credit from Form 3800"
    reference = "https://www.law.cornell.edu/uscode/text/26/38"
    unit = USD



from policyengine_us.model_api import *


class mt_federal_income_tax_deduction_unit(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Montana federal income tax deduction for the entire tax unit"
    reference = (
        "https://law.justia.com/codes/montana/2021/title-15/chapter-30/part-21/section-15-30-2131/"
        # MT Code § 15-30-2131 (2021) (1)(b)
    )
    unit = USD
    defined_for = StateCode.MT

    def formula(tax_unit, period, parameters):
        filing_status = tax_unit("filing_status", period)
        p = parameters(
            period
        ).gov.states.mt.tax.income.deductions.itemized.federal_income_tax
        federal_income_tax_limit = p.cap[filing_status]

        federal_income_tax = tax_unit(
            "income_tax_before_refundable_credits", period
        )
        # The federal income tax deduction is attributed to the head in any case
        # as we currently do not have a way to determine individual federal income tax before credits.
        return min_(federal_income_tax, federal_income_tax_limit)



from policyengine_us.model_api import *


class sc_young_child_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "South Carolina young child deduction"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://dor.sc.gov/forms-site/Forms/SC1040_2022.pdf#page=2",
        "https://www.scstatehouse.gov/code/t12c006.php",
        # SECTION 12-6-1160
    )
    defined_for = StateCode.SC

    def formula(tax_unit, period, parameters):
        # Get relevant parameter subtree.
        p = parameters(period).gov.states.sc.tax.income.deductions.young_child
        # Determine eligibility for each person in the tax unit.
        person = tax_unit.members
        meets_age_limit = person("age", period) < p.ineligible_age
        eligible_dependent = meets_age_limit & person(
            "is_tax_unit_dependent", period
        )
        # Count number of eligible people in the tax unit.
        total_eligible_dependents = tax_unit.sum(eligible_dependent)
        # Multiply by the amount per eligible dependent.
        return total_eligible_dependents * p.amount



from policyengine_us.model_api import *


class nc_standard_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "North Carolina standard deduction"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.NC

    def formula(tax_unit, period, parameters):
        filing_status = tax_unit("filing_status", period)
        p = parameters(period).gov.states.nc.tax.income
        return p.deductions.standard.amount[filing_status]



from policyengine_us.model_api import *


class ut_total_dependents(Variable):
    value_type = int
    entity = TaxUnit
    label = "Utah total dependents"
    unit = USD
    documentation = "Form TC-40, line 2c"
    definition_period = YEAR
    defined_for = StateCode.UT
    adds = ["tax_unit_count_dependents"]



from policyengine_us.model_api import *


class is_snap_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "SNAP eligible"
    documentation = "Whether this SPM unit is eligible for SNAP benefits"
    definition_period = MONTH
    reference = (
        "https://www.law.cornell.edu/uscode/text/7/2017#a",
        "https://www.law.cornell.edu/uscode/text/7/2014#c",
    )

    def formula(spm_unit, period, parameters):
        net = spm_unit("meets_snap_net_income_test", period)
        gross = spm_unit("meets_snap_gross_income_test", period)
        asset = spm_unit("meets_snap_asset_test", period)
        normal_eligibility = net & gross & asset
        # Categorical eligibility (SSI, TANF, and BBCE TANF) overrides tests.
        categorical_eligibility = spm_unit(
            "meets_snap_categorical_eligibility", period
        )
        person = spm_unit.members
        eligible_person_present = spm_unit.any(
            ~person("is_snap_ineligible_student", period)
        )
        work_requirements_eligibility = spm_unit(
            "meets_snap_work_requirements", period
        )
        return (
            (normal_eligibility | categorical_eligibility)
            & eligible_person_present
            & work_requirements_eligibility
        )



from policyengine_us.model_api import *


class is_ssi_qualified_noncitizen(Variable):
    value_type = bool
    entity = Person
    label = "Is an SSI qualified noncitizen"
    definition_period = YEAR
    reference = "https://secure.ssa.gov/poms.nsf/lnx/0500502100"

    def formula(person, period, parameters):
        qualifying_quarters_earnings = person(
            "ssi_qualifying_quarters_earnings", period
        )
        p = parameters(period).gov.ssa.ssi
        immigration_status = person("immigration_status", period)
        legal_permanent_resident = (
            immigration_status
            == immigration_status.possible_values.LEGAL_PERMANENT_RESIDENT
        )
        # For legal permanent residents, check if they have enough qualifying quarters
        meets_earnings_requirement = (
            qualifying_quarters_earnings
            >= p.income.sources.qualifying_quarters_threshold
        )

        # Non-LPRs automatically meet the earnings requirement
        # LPRs must meet the earnings threshold
        earnings_requirement_satisfied = (
            ~legal_permanent_resident
        ) | meets_earnings_requirement

        # Convert immigration status to string and check if it's in the list of qualified statuses
        immigration_status_str = immigration_status.decode_to_str()
        has_qualifying_status = np.isin(
            immigration_status_str,
            p.eligibility.status.qualified_noncitizen_status,
        )
        return has_qualifying_status & earnings_requirement_satisfied



from policyengine_us.model_api import *


class mi_disabled_exemption_eligible_person(Variable):
    value_type = float
    entity = Person
    label = "Eligible person for the Michigan disabled exemptions"
    defined_for = StateCode.MI
    unit = USD
    definition_period = YEAR
    reference = (
        "http://legislature.mi.gov/doc.aspx?mcl-206-30",
        "https://www.legislature.mi.gov/Publications/TaxpayerGuide.pdf",
    )

    def formula(person, period, parameters):
        blind = person("is_blind", period)
        deaf = person("is_deaf", period)
        # Only totally disabled people under a certain age are eligible
        totally_disabled = person(
            "is_permanently_and_totally_disabled", period
        )
        p = parameters(period).gov.states.mi.tax.income.exemptions.disabled
        age_eligible = person("age", period) < p.age_limit
        totally_disabled_eligible = totally_disabled & age_eligible

        return blind | deaf | totally_disabled_eligible



from policyengine_us.model_api import *
from policyengine_us.variables.gov.ssa.ssi.eligibility.income._apply_ssi_exclusions import (
    _apply_ssi_exclusions,
)


class ssi_ineligible_parent_allocation(Variable):
    value_type = float
    entity = Person
    label = "SSI ineligible parent allocation"
    unit = USD
    documentation = "The amount of income that SSI does not to SSI claimants."
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/cfr/text/20/416.1163"

    def formula(person, period, parameters):
        ssi = parameters(period).gov.ssa.ssi.amount
        ineligible_parent = person("is_ssi_ineligible_parent", period)
        count_ineligible_parents = person.tax_unit.sum(ineligible_parent)
        return (
            where(
                count_ineligible_parents == 2,
                ssi.couple / 2,
                ssi.individual,
            )
            * MONTHS_IN_YEAR
            * ineligible_parent
        )



from policyengine_us.model_api import *


class ctc_arpa_addition(Variable):
    value_type = float
    entity = TaxUnit
    label = "Additional CTC from ARPA"
    unit = USD
    definition_period = YEAR
    # Defined on Line 5 worksheet of 2021 Instructions for Schedule 8812.
    reference = "https://www.irs.gov/pub/irs-pdf/i1040s8.pdf#page=4"
    adds = ["ctc_arpa_max_addition"]
    subtracts = ["ctc_arpa_phase_out"]



from policyengine_us.model_api import *


class oh_income_tax_exempt(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Ohio income tax exempt"
    defined_for = StateCode.OH
    unit = USD
    definition_period = YEAR
    reference = "https://tax.ohio.gov/static/forms/ohio_individual/individual/2021/pit-it1040-booklet.pdf"

    def formula(tax_unit, period, parameters):
        taxable_income = tax_unit("oh_taxable_income", period)
        p = parameters(period).gov.states.oh.tax.income
        return taxable_income < p.agi_threshold



from policyengine_us.model_api import *


class md_dependent_care_subtraction(Variable):
    value_type = float
    entity = TaxUnit
    label = "MD depdendent care subtraction from AGI"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://casetext.com/statute/code-of-maryland/article-tax-general/title-10-income-tax/subtitle-2-maryland-taxable-income-calculations-for-individual/part-ii-maryland-adjusted-gross-income/section-10-208-effective-until-712024-subtractions-from-federal-adjusted-gross-income-state-adjustments"
        "https://www.marylandtaxes.gov/forms/21_forms/Resident_Booklet.pdf#page=13"
        "https://www.marylandtaxes.gov/forms/21_forms/Resident_Booklet.pdf#page=13"
    )
    defined_for = StateCode.MD

    """
    (e) The subtraction under subsection (a) of this section includes expenses
    for household and dependent care services not exceeding the dollar limit
    allowed under § 21(c) of the Internal Revenue Code and determined without
    reference to the percentage limitation in § 21(a)(2) of the Internal
    Revenue Code.

    IRC § 21(c): https://www.law.cornell.edu/uscode/text/26/21#c
    IRC § 21(a)(2): https://www.law.cornell.edu/uscode/text/26/21#a_2
    """

    def formula(tax_unit, period, parameters):
        md_p = parameters(period).gov.states.md.tax.income.agi.subtractions
        max_decoupled_year_offset = md_p.max_care_expense_year_offset
        period_max = period.offset(max_decoupled_year_offset)
        md_max_care_expense = parameters(period_max).gov.irs.credits.cdcc.max
        max_eligibles = parameters(period).gov.irs.credits.cdcc.eligibility.max
        num_eligibles = min_(
            max_eligibles, tax_unit("count_cdcc_eligible", period)
        )
        us_expenses = tax_unit("cdcc_relevant_expenses", period)
        return min_(md_max_care_expense * num_eligibles, us_expenses)



from policyengine_us.model_api import *


class tax_exempt_retirement_distributions(Variable):
    value_type = float
    entity = Person
    label = "Tax-exempt retirement account distributions"
    unit = USD
    definition_period = YEAR
    adds = [
        "tax_exempt_ira_distributions",
        "tax_exempt_401k_distributions",
        "tax_exempt_sep_distributions",
        "tax_exempt_403b_distributions",
    ]



from policyengine_us.model_api import *


class vt_cdcc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Vermont child care and dependent care credit"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://tax.vermont.gov/sites/tax/files/documents/IN-112-2022.pdf#page=2"
        "https://law.justia.com/codes/vermont/2022/title-32/chapter-151/section-5828c/"
        "https://www.irs.gov/pub/irs-prior/f2441--2022.pdf#page=1"
    )
    defined_for = StateCode.VT

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.vt.tax.income.credits.cdcc
        # The form refers to 2022 Form 2441 line 11, which caps the credit at tax liability.
        federal_cdcc = tax_unit("capped_cdcc", period)
        return federal_cdcc * p.rate



from policyengine_us.model_api import *


class care_expenses(Variable):
    value_type = float
    entity = Person
    label = "Care expenses"
    unit = USD
    definition_period = MONTH

    def formula(person, period, parameters):
        return person("pre_subsidy_care_expenses", period)

    # Add subsidies in a .yaml file once added to the model



from policyengine_us.model_api import *


class md_married_or_has_child_refundable_eitc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Maryland refundable EITC for filers who are married or have qualifying child"
    unit = USD
    definition_period = YEAR
    reference = "https://www.marylandtaxes.gov/forms/21_forms/Resident_Booklet.pdf#page=23"
    defined_for = StateCode.MD

    def formula(tax_unit, period, parameters):
        # Limited to filers who are married or have child
        does_not_qualify_for_unmarried_childless_eitc = ~tax_unit(
            "md_qualifies_for_unmarried_childless_eitc", period
        )

        federal_eitc = tax_unit("eitc", period)
        md_tax = tax_unit("md_income_tax_before_credits", period)
        p = parameters(
            period
        ).gov.states.md.tax.income.credits.eitc.refundable.married_or_has_child

        total_refundable_eitc_amount = federal_eitc * p.match
        applicable_refundable_eitc = max_(
            total_refundable_eitc_amount - md_tax, 0
        )

        return (
            does_not_qualify_for_unmarried_childless_eitc
            * applicable_refundable_eitc
        )



from policyengine_us.model_api import *


class id_retirement_benefits_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Idaho retirement benefits deduction"
    unit = USD
    reference = [
        "https://legislature.idaho.gov/statutesrules/idstat/title63/t63ch30/sect63-3022a/",
        "https://tax.idaho.gov/wp-content/uploads/forms/EFO00088/EFO00088_03-01-2023.pdf",
    ]
    definition_period = YEAR
    defined_for = StateCode.ID

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.id.tax.income.deductions.retirement_benefits
        filing_status = tax_unit("filing_status", period)
        # Line 8a
        # Max retirement benefits deduction amount
        cap = p.cap[filing_status]
        # Line 8c
        # Social Security retirement benefits received
        person = tax_unit.members
        head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        ss_retirement = person("social_security_retirement", period)
        total_head_spouse_ss = tax_unit.sum(ss_retirement * head_or_spouse)
        # Line 8d
        # Base amount minus social Security benefits received
        ded_amt = max_(cap - total_head_spouse_ss, 0)
        # Line 8e
        # Qualified retirement benefits included in federal income
        # The head or spouse condition is included in the eligible person variable
        relevant_income = add(
            tax_unit,
            period,
            ["id_retirement_benefits_deduction_relevant_income"],
        )
        # Line 8f
        return min_(ded_amt, relevant_income)



from policyengine_us.model_api import *


class tx_ccs_eligible_child(Variable):
    value_type = bool
    entity = Person
    label = "Texas CCS eligible child"
    definition_period = MONTH
    reference = "https://www.law.cornell.edu/regulations/texas/40-Tex-Admin-Code-SS-809-41"
    defined_for = StateCode.TX

    def formula(person, period, parameters):
        p = parameters(period).gov.states.tx.twc.ccs.age_threshold
        age = person("age", period.this_year)
        is_disabled = person("is_disabled", period)
        age_limit = where(is_disabled, p.disabled_child, p.child)
        age_eligible = age < age_limit
        is_dependent = person("is_tax_unit_dependent", period)
        immigration_status_eligible = person(
            "is_citizen_or_legal_immigrant", period
        )
        return age_eligible & is_dependent & immigration_status_eligible



from policyengine_us.model_api import *


class nc_tanf_income_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "North Carolina TANF income eligible"
    definition_period = YEAR
    defined_for = "nc_demographic_tanf_eligible"

    def formula(spm_unit, period, parameters):
        monthly_allowed_difference = parameters(
            period
        ).gov.states.nc.ncdhhs.tanf.need_standard.average_reduced_need_standard_thresold
        household_size = spm_unit("nc_tanf_household_size", period)
        reduced_need_standard = spm_unit(
            "nc_tanf_reduced_need_standard", period
        )

        need_standard_fraction = reduced_need_standard / household_size
        difference_threshold = monthly_allowed_difference * MONTHS_IN_YEAR

        return need_standard_fraction >= difference_threshold



from policyengine_us.model_api import *


class taxable_unemployment_compensation(Variable):
    value_type = float
    entity = Person
    label = "Taxable unemployment compensation"
    unit = USD
    definition_period = YEAR

    def formula(person, period, parameters):
        # The taxable amount of unemployment compensation is decided at the tax unit level, but
        # gross income (which contains taxable UI) is person-level. Therefore, we include the
        # taxable UI in gross income by assigning it to the head of the tax unit: this will be
        # not affect overall tax liability.

        is_tax_unit_head = person("is_tax_unit_head", period)
        tax_unit_taxable_uc = person.tax_unit(
            "tax_unit_taxable_unemployment_compensation", period
        )
        return where(is_tax_unit_head, tax_unit_taxable_uc, 0)



from policyengine_us.model_api import *


class ma_liheap_income(Variable):
    value_type = float
    entity = SPMUnit
    label = "Massachusetts LIHEAP income"
    definition_period = YEAR
    defined_for = StateCode.MA
    reference = "https://www.mass.gov/info-details/learn-about-home-energy-assistance-heap"

    # The income concept is not clearly defined, assuming IRS gross income
    adds = ["irs_gross_income"]



from policyengine_us.model_api import *


class is_ccdf_home_based(Variable):
    value_type = bool
    default_value = False
    entity = Person
    label = "Whether CCDF care is home-based versus center-based"
    definition_period = YEAR

    def formula(person, period, parameters):
        provider_type_group = person("childcare_provider_type_group", period)
        provider_type_groups = provider_type_group.possible_values
        return provider_type_group != provider_type_groups.DCC_SACC



from policyengine_us.model_api import *


class ky_service_credit_months_post_1997(Variable):
    value_type = float
    entity = Person
    label = "Kentucky service credit months after 1997"
    definition_period = YEAR
    defined_for = StateCode.KY



from policyengine_us.model_api import *


class co_ccap_re_determination_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Eligible for the re-determination of the Colorado Child Care Assistance Program"
    reference = "https://www.sos.state.co.us/CCR/GenerateRulePdf.do?ruleVersionId=11042&fileName=8%20CCR%201403-1#page=31"
    definition_period = MONTH
    # defined_for = StateCode.CO

    def formula(spm_unit, period, parameters):
        income_eligible = spm_unit(
            "co_ccap_re_determination_income_eligible", period
        )
        has_eligible_children = (
            spm_unit("co_ccap_eligible_children", period) > 0
        )
        return income_eligible & has_eligible_children



from policyengine_us.model_api import *


class taxsim_sage(Variable):
    value_type = int
    entity = TaxUnit
    label = "Age of spouse"
    unit = "year"
    documentation = "Age of spouse (zero if unknown or not a joint return). It is an error to specify a non-zero spouse age for an unmarried taxpayer."
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        is_head = person("is_tax_unit_spouse", period)
        age = person("age", period)
        return tax_unit.sum(age * is_head)



from policyengine_us.model_api import *


class ma_liheap(Variable):
    value_type = float
    entity = SPMUnit
    label = "Massachusetts LIHEAP payment"
    definition_period = YEAR
    defined_for = "ma_liheap_eligible"
    reference = "https://www.mass.gov/doc/fy-2025-heap-income-eligibility-benefit-chart-may-8-2025/download"

    def formula(spm_unit, period, parameters):
        payment_amount = add(
            spm_unit,
            period,
            ["ma_liheap_standard_payment", "ma_liheap_hecs_payment"],
        )
        actual_expense_amount = add(
            spm_unit,
            period,
            ["heating_cooling_expense", "gas_expense", "electricity_expense"],
        )
        return min_(actual_expense_amount, payment_amount)



from policyengine_us.model_api import *


class md_tanf_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Maryland TANF eligible"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MD



from policyengine_us.model_api import *


class oh_adoption_credit_person(Variable):
    value_type = float
    entity = Person
    label = "Ohio adoption credit for each person"
    unit = USD
    definition_period = YEAR
    reference = (
        # Ohio 2022 Instructions for Filing Original and Amended - Line 18 – Ohio Adoption Credit
        "https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf#page=21",
        # Ohio Income - Individual Credits (Education, Displaced Workers & Adoption) 15. - 20.
        "https://tax.ohio.gov/wps/portal/gov/tax/help-center/faqs/income+-+individual+credits/income-individual-credits",
        # The ohio adoption credit has been repealed in the newest revised code, but can still be found in the previous legal code and 2022 tax form
        # 2023 Ohio Rev. Code § 5747.37
        "https://casetext.com/statute/ohio-revised-code/title-57-taxation/chapter-5747-income-tax/section-574737-repealed",
        # 2022 Ohio Revised Code Title 57 | Taxation Chapter 5747 | Income Tax Section 5747.37 | Credit for Legally Adopted Minor Child.
        "https://law.justia.com/codes/ohio/2022/title-57/chapter-5747/section-5747-37/",
    )
    defined_for = StateCode.OH

    def formula(person, period, parameters):
        eligible_adoption_related_expenses = person(
            "qualified_adoption_assistance_expense", period
        )
        p = parameters(period).gov.states.oh.tax.income.credits.adoption
        child_age_eligible = person("age", period) < p.age_limit
        adopted_this_year = person("adopted_this_year", period)
        age_eligible_adopted_child = child_age_eligible & adopted_this_year
        credit_if_eligible = min_(
            max_(eligible_adoption_related_expenses, p.amount.min),
            p.amount.max,
        )
        return credit_if_eligible * age_eligible_adopted_child



from policyengine_us.model_api import *


class prior_year_minimum_tax_credit(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    label = "Prior year minimum tax credit"
    documentation = "Prior year minimum tax credit from Form 8801"
    reference = "https://www.law.cornell.edu/uscode/text/26/53"
    unit = USD



from policyengine_us.model_api import *


class count_cdcc_eligible(Variable):
    value_type = int
    entity = TaxUnit
    label = "CDCC-eligible children"
    unit = USD
    definition_period = YEAR

    adds = ["is_cdcc_eligible"]



from policyengine_us.model_api import *


class tax_unit_partnership_s_corp_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "Tax unit partnership/S-corp income"
    unit = USD
    documentation = (
        "Combined partnership/S-corporation income for the tax unit."
    )
    definition_period = YEAR

    adds = ["partnership_s_corp_income"]



from policyengine_us.model_api import *


class md_aged_blind_exemptions(Variable):
    value_type = float
    entity = TaxUnit
    label = "MD aged blind exemptions"
    unit = USD
    definition_period = YEAR
    reference = "https://govt.westlaw.com/mdc/Document/NF59A76006EA511E8ABBEE50DE853DFF4?viewType=FullText&originationContext=documenttoc&transitionType=CategoryPageItem&contextData=(sc.Default)"
    defined_for = StateCode.MD

    adds = [
        "md_aged_dependent_exemption",
        "md_aged_exemption",
        "md_blind_exemption",
    ]



from policyengine_us.model_api import *


class ct_taxable_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "Connecticut taxable income"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.CT

    def formula(tax_unit, period, parameters):
        income = tax_unit("ct_agi", period)
        personal_exemptions = tax_unit("ct_personal_exemptions", period)
        return max_(income - personal_exemptions, 0)



from policyengine_us.model_api import *


class mi_interest_dividends_capital_gains_deduction_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = (
        "Eligible for the Michigan interest dividends capital gains deduction"
    )
    definition_period = YEAR
    reference = (
        "http://legislature.mi.gov/doc.aspx?mcl-206-30",  # (1)(p)
        "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf#page=16",
    )
    defined_for = StateCode.MI

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.mi.tax.income.deductions.interest_dividends_capital_gains

        older_spouse_birth_year = tax_unit("older_spouse_birth_year", period)

        return older_spouse_birth_year < p.birth_year



from policyengine_us.model_api import *


class ma_eaedc_immigration_status_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Eligible for the Massachusetts EAEDC based on immigration status"
    definition_period = MONTH
    reference = (
        "https://www.law.cornell.edu/regulations/massachusetts/106-CMR-703-440"
    )
    defined_for = StateCode.MA

    def formula(spm_unit, period, parameters):
        person = spm_unit.members
        head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        immigration_status = person("immigration_status", period)
        is_undocumented = (
            immigration_status
            == immigration_status.possible_values.UNDOCUMENTED
        )
        return spm_unit.any(head_or_spouse & ~is_undocumented)



from policyengine_us.model_api import *


class la_general_relief_immigration_status_eligible_person(Variable):
    value_type = bool
    entity = Person
    definition_period = YEAR
    label = "Eligible Person for the Los Angeles County General Relief based on the immigration status requirements"
    # Person has to be a resident of LA County
    defined_for = "in_la"
    reference = "http://epolicy.dpss.lacounty.gov/epolicy/epolicy/server/general/projects_responsive/ePolicyMaster/index.htm?&area=general&type=responsivehelp&ctxid=&project=ePolicyMaster#t=mergedProjects%2FGR%2FGR%2F42-404_Immigrant_Eligibility_Chart%2F42-404_Immigrant_Eligibility_Chart.htm"

    def formula(person, period, parameters):
        # Undocuemnted as well as DACA classified applicants/participants are ineligible for the GR
        istatus = person("immigration_status", period)
        daca_tps = istatus == istatus.possible_values.DACA_TPS
        undocumented = istatus == istatus.possible_values.UNDOCUMENTED
        head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        # Assuming that the applicant's/participant's immigration status is recorded
        return ~(daca_tps | undocumented) & head_or_spouse



from policyengine_us.model_api import *


class ctc_qualifying_children(Variable):
    value_type = int
    entity = TaxUnit
    label = "CTC-qualifying children"
    documentation = "Count of children that qualify for the Child Tax Credit"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/24#c"

    adds = ["ctc_qualifying_child"]



from policyengine_us.model_api import *


class ca_riv_share_countable_income(Variable):
    value_type = float
    entity = SPMUnit
    unit = USD
    label = "Riverside County Sharing Households Assist Riverside's Energy program (SHARE) countable income"
    definition_period = YEAR
    defined_for = "in_riv"
    reference = "https://riversideca.gov/utilities/residents/assistance-programs/share-english"

    adds = "gov.local.ca.riv.cap.share.countable_income.sources"



from policyengine_us.model_api import *


class snap_limited_utility_allowance_by_household_size(Variable):
    value_type = bool
    entity = SPMUnit
    label = "SNAP Limited Utility Allowance by household size"
    unit = USD
    definition_period = MONTH

    def formula(spm_unit, period, parameters):
        p = parameters(period).gov.usda.snap.income.deductions.utility.limited
        region = spm_unit.household("snap_utility_region_str", period)
        return is_in(region, p.by_household_size.participating_states)



from policyengine_us.model_api import *


class tx_tanf_income_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Meets Texas TANF income test"
    definition_period = MONTH
    reference = "https://www.hhs.texas.gov/handbooks/texas-works-handbook/a-1340-income-limits"
    defined_for = StateCode.TX

    def formula(spm_unit, period, parameters):
        # Texas TANF has two income eligibility tests per § 372.408:
        #
        # Budgetary needs test (§ 372.408 (a)(1)):
        #   - Uses income WITHOUT earned income disregards (1/3 or 90%)
        #   - Income < budgetary needs amount (100% of needs)
        #   - Applies to: NEW applicants only
        #
        # Recognizable needs test (§ 372.408 (a)(2)):
        #   - Uses income WITH earned income disregards (1/3 or 90%)
        #   - Income < recognizable needs amount (25% of budgetary needs)
        #   - Applies to: EVERYONE (applicants AND continuing recipients)
        #
        # Test requirements by enrollment status:
        # - New applicants: Must pass budgetary test AND recognizable test
        # - Continuing recipients: Must pass recognizable test only

        is_enrolled = spm_unit("is_tanf_enrolled", period)
        passes_budgetary = spm_unit("tx_tanf_budgetary_needs_test", period)
        passes_recognizable = spm_unit(
            "tx_tanf_recognizable_needs_test", period
        )

        # Apply the appropriate test(s) based on enrollment status
        return where(
            is_enrolled,
            passes_recognizable,  # Continuing: recognizable test only
            passes_budgetary & passes_recognizable,  # Applicants: BOTH tests
        )



from policyengine_us.model_api import *


class in_qualifying_child_count(Variable):
    value_type = int
    entity = TaxUnit
    label = "Indiana qualifying dependent child count"
    unit = "child"
    documentation = (
        "Number of qualifying children for the IN additional exemption."
    )
    definition_period = YEAR
    reference = "http://iga.in.gov/legislative/laws/2021/ic/titles/006#6-3-1-3.5"  # (5)(B)(i)
    defined_for = StateCode.IN

    adds = ["in_is_qualifying_dependent_child"]



from policyengine_us.model_api import *


class short_term_capital_gains(Variable):
    value_type = float
    entity = Person
    label = "short-term capital gains"
    unit = USD
    documentation = "Net gains made from sales of assets held for one year or less(losses are expressed as negative gains)."
    definition_period = YEAR
    reference = dict(
        title="26 U.S. Code § 1222(1)",
        href="https://www.law.cornell.edu/uscode/text/26/1222#1",
    )
    uprating = "calibration.gov.irs.soi.short_term_capital_gains"



from policyengine_us.model_api import *


class tx_tanf_earned_income_after_disregard_person(Variable):
    value_type = float
    entity = Person
    label = "Texas TANF earned income after disregards (person)"
    unit = USD
    definition_period = MONTH
    reference = (
        "https://www.hhs.texas.gov/handbooks/texas-works-handbook/a-1420-types-deductions",
        "https://www.law.cornell.edu/regulations/texas/1-Tex-Admin-Code-SS-372-409",
    )
    defined_for = StateCode.TX

    def formula(person, period, parameters):
        # Get gross earned income
        gross_earned = person("tx_tanf_gross_earned_income", period)
        p = parameters(period).gov.states.tx.tanf.income

        # Step 1: Subtract work expense deduction ($120, not to exceed earnings)
        work_expense = min_(p.deductions.work_expense, gross_earned)
        after_work_expense = max_(gross_earned - work_expense, 0)

        # Step 2: Apply appropriate earned income disregard
        # Check if household is enrolled in TANF
        spm_unit = person.spm_unit
        is_enrolled = spm_unit("is_tanf_enrolled", period)

        # For applicants (not enrolled): 1/3 disregard
        applicant_disregard = (
            after_work_expense * p.disregards.applicant_fraction
        )

        # For continuing recipients (enrolled): 90% disregard (capped at $1,400)
        # Note: This disregard is limited to 4 months per 12-month period (not yet implemented)
        potential_recipient_disregard = (
            after_work_expense * p.disregards.continuing_recipient_rate
        )
        actual_recipient_disregard = min_(
            potential_recipient_disregard,
            p.disregards.continuing_recipient_cap,
        )

        # Return appropriate value based on enrollment status
        disregard = where(
            is_enrolled,
            actual_recipient_disregard,
            applicant_disregard,
        )

        return max_(after_work_expense - disregard, 0)



from policyengine_us.model_api import *


class spm_unit_net_income(Variable):
    value_type = float
    entity = SPMUnit
    label = "Net income"
    definition_period = YEAR
    unit = USD

    adds = ["spm_unit_market_income", "spm_unit_benefits"]
    subtracts = ["spm_unit_taxes", "spm_unit_spm_expenses"]



from policyengine_us.model_api import *


class hi_food_excise_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Hawaii Food/Excise Tax Credit"
    defined_for = StateCode.HI
    unit = USD
    definition_period = YEAR
    reference = "https://files.hawaii.gov/tax/legal/hrs/hrs_235.pdf#page=44"

    def formula(tax_unit, period, parameters):
        # filer can not be a dependent on another return
        dependent_elsewhere = tax_unit("head_is_dependent_elsewhere", period)
        total_amount = add(
            tax_unit,
            period,
            [
                "hi_food_excise_credit_minor_child_amount",
                "hi_food_excise_exemption_amount",
            ],
        )
        return ~dependent_elsewhere * total_amount



from policyengine_us.model_api import *


class il_aabd_eligible_person(Variable):
    value_type = bool
    entity = Person
    definition_period = MONTH
    label = "Eligible person for Illinois Aid to the Aged, Blind or Disabled (AABD)"
    defined_for = StateCode.IL

    def formula(person, period, parameters):
        financial_eligible = person(
            "il_aabd_financial_eligible_person", period
        )
        non_financial_eligible = person(
            "il_aabd_non_financial_eligible_person", period
        )
        return financial_eligible & non_financial_eligible



from policyengine_us.model_api import *


class relative_wage_change(Variable):
    value_type = float
    entity = Person
    label = "relative wage change"
    unit = "/1"
    definition_period = YEAR
    requires_computation_after = "labor_supply_behavioral_response"

    def formula(person, period, parameters):
        simulation = person.simulation
        baseline_branch = simulation.get_branch("baseline").get_branch(
            "baseline_lsr_measurement"
        )
        baseline_person = baseline_branch.populations["person"]
        baseline_mtr = baseline_person("marginal_tax_rate", period)
        baseline_wage = 1 - baseline_mtr
        measurement_branch = simulation.get_branch("lsr_measurement")
        measurement_person = measurement_branch.populations["person"]
        mtr = measurement_person("marginal_tax_rate", period)
        wage_rate = 1 - mtr
        # _c suffix for "clipped"
        baseline_wage_c = np.where(baseline_wage == 0, 0.01, baseline_wage)
        wage_rate_c = np.where(wage_rate == 0, 0.01, wage_rate)
        relative_change = (wage_rate_c - baseline_wage_c) / baseline_wage_c
        p = parameters(period).gov.simulation.labor_supply_responses
        return np.clip(
            relative_change,
            -p.bounds.effective_wage_rate_change,
            p.bounds.effective_wage_rate_change,
        )



from policyengine_us.model_api import *


class in_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "Indiana refundable income tax credits"
    unit = USD
    definition_period = YEAR
    reference = "https://iga.in.gov/laws/2021/ic/titles/6#6-3.1"
    defined_for = StateCode.IN
    # Use this instead of the parameter because the .in breaks the adds.
    # Use this instead of a formula so the app displays the breakdown.
    adds = ["in_eitc", "in_unified_elderly_tax_credit"]



from policyengine_us.model_api import *


class greater_age_head_spouse(Variable):
    value_type = int
    entity = TaxUnit
    definition_period = YEAR
    label = "Age of head or spouse of tax unit depending on which is greater"
    documentation = "Age in years of taxpayer (i.e. primary adult) or spouse (i.e. secondary adult if present), depending on which is greater. "
    unit = "year"

    def formula(tax_unit, period, parameters):
        return max_(
            tax_unit("age_head", period), tax_unit("age_spouse", period)
        )



from policyengine_us.model_api import *


class ca_foster_youth_tax_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "California foster youth tax credit"
    unit = USD
    definition_period = YEAR
    defined_for = "ca_eitc_eligible"

    adds = ["ca_foster_youth_tax_credit_person"]



from policyengine_us.model_api import *


class hi_itemized_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Hawaii itemized deduction"
    unit = USD
    documentation = (
        "https://files.hawaii.gov/tax/forms/2022/n11ins.pdf#page=15"
        "https://files.hawaii.gov/tax/forms/2022/n11ins.pdf#page=32"  # total itemized deduction worksheet
    )
    definition_period = YEAR
    defined_for = StateCode.HI

    def formula(tax_unit, period, parameters):
        # Hawaii applies an federal AGI limit which has been introduced in 2009
        p = parameters(f"2009-01-01").gov.irs.deductions.itemized.reduction

        # Hawaii did not suspend the overall limitation on itemized deductions
        # You may not be able to deduct all of your itemized deductions if agi reach the cap
        # need to calculate the reduced itemized deductions
        hi_agi = tax_unit("hi_agi", period)
        filing_status = tax_unit("filing_status", period)
        total_itemized_deduction_eligibility = (
            hi_agi < p.agi_threshold[filing_status]
        )
        return where(
            total_itemized_deduction_eligibility,
            tax_unit("hi_total_itemized_deductions", period),
            tax_unit("hi_reduced_itemized_deductions", period),
        )



from policyengine_us.model_api import *


class wi_capital_gain_loss_subtraction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Wisconsin capital gain/loss subtraction from federal AGI"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.revenue.wi.gov/TaxForms2021/2021-ScheduleWDf.pdf#page=2"
        "https://www.revenue.wi.gov/TaxForms2022/2022-ScheduleWDf.pdf#page=2"
    )
    defined_for = StateCode.WI

    def formula(tax_unit, period, parameters):
        # calculate Schedule WD, Line 8
        stcg_net = add(tax_unit, period, ["short_term_capital_gains"])
        # calculate Schedule WD, Line 17
        ltcg_net = add(tax_unit, period, ["long_term_capital_gains"])
        # calculate Schedule WD, Line 18
        totcg = max_(0, stcg_net + ltcg_net)
        # calculate Schedule WD, Line 20, the capital gain reduction
        p = parameters(period).gov.states.wi.tax.income.subtractions
        fraction = p.capital_gain.fraction
        cg_reduction = min_(totcg, max_(0, ltcg_net)) * fraction
        # calculate Schedule WD, Line 27, the WI reduced capital gain
        wi_cg = totcg - cg_reduction
        # calculate Schedule WD, Line 29a
        us_cg = totcg
        # return Schedule SB capital gain subtraction (WB Line 29d)
        return max_(0, us_cg - wi_cg)



from policyengine_us.model_api import *


class il_liheap_base_payment(Variable):
    value_type = float
    entity = SPMUnit
    label = "Illinois LIHEAP base payment"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.IL
    reference = "https://dceo.illinois.gov/communityservices/utilitybillassistance.html"

    def formula(spm_unit, period, parameters):
        p = parameters(period).gov.states.il.dceo.liheap.payment.base_amount

        # Check if heat is included in rent
        heat_in_rent = spm_unit("heat_expense_included_in_rent", period)
        heating_expenses = add(spm_unit, period, ["heating_expense_person"])

        # For renters with heat included in rent, provide minimum cash benefit
        # For others, benefit based on actual heating expenses up to maximum
        capped_heating_expenses = min_(heating_expenses, p.max)

        return where(
            heat_in_rent,
            p.min,  # Fixed minimum for heat included in rent
            capped_heating_expenses,  # Based on actual expenses, up to maximum
        )



from policyengine_us.model_api import *


class MassachusettsLIHEAPUtilityCategory(Enum):
    UTILITY_AND_HEAT_IN_RENT = (
        "Utility and Heat in Rent"  # Electric, Natural Gas, Heat in Rent
    )
    DELIVERABLE_FUEL = (
        "Deliverable Fuel"  # Kerosene, Heating Oil, Propane, Other
    )
    NONE = "None"


class ma_liheap_utility_category(Variable):
    value_type = Enum
    entity = SPMUnit
    possible_values = MassachusettsLIHEAPUtilityCategory
    default_value = MassachusettsLIHEAPUtilityCategory.NONE
    label = "Massachusetts LIHEAP household's utility category"
    definition_period = YEAR
    defined_for = StateCode.MA
    reference = "https://www.mass.gov/doc/fy-2025-heap-income-eligibility-benefit-chart-may-8-2025/download"

    def formula(spm_unit, period, parameters):
        heat_in_rent = spm_unit("heat_expense_included_in_rent", period)
        heating_type = spm_unit("ma_liheap_heating_type", period)
        heating_types = heating_type.possible_values

        electricity = heating_type == heating_types.ELECTRICITY
        natural_gas = heating_type == heating_types.NATURAL_GAS
        heating_and_oil_propane = (
            heating_type == heating_types.HEATING_OIL_AND_PROPANE
        )
        kerosene = heating_type == heating_types.KEROSENE
        other = heating_type == heating_types.OTHER

        gas_or_electric_heating = electricity | natural_gas
        utility_and_heat_in_rent = gas_or_electric_heating | heat_in_rent

        deliverable_fuel = heating_and_oil_propane | kerosene | other

        conditions = [
            utility_and_heat_in_rent,
            deliverable_fuel,
        ]
        results = [
            MassachusettsLIHEAPUtilityCategory.UTILITY_AND_HEAT_IN_RENT,
            MassachusettsLIHEAPUtilityCategory.DELIVERABLE_FUEL,
        ]

        return select(
            conditions,
            results,
            default=MassachusettsLIHEAPUtilityCategory.NONE,
        )



from policyengine_us.model_api import *


class ma_withheld_income_tax(Variable):
    value_type = float
    entity = Person
    label = "Massachusetts withheld income tax"
    defined_for = StateCode.MA
    unit = USD
    definition_period = YEAR

    def formula(person, period, parameters):
        agi = person("adjusted_gross_income_person", period)
        p = parameters(period).gov.states.ma.tax.income
        # Since Massachusetts does not have a standard deduction, we apply the
        # base personal exemption amount
        personal_exemption = p.exemptions.personal["SINGLE"]
        reduced_agi = max_(agi - personal_exemption, 0)
        return p.rates.part_b * reduced_agi



from policyengine_us.model_api import *


class ks_fstc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Kansas food sales tax credit"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.ksrevenue.gov/pdf/ip21.pdf"
        "https://www.ksrevenue.gov/pdf/ip22.pdf"
    )
    defined_for = StateCode.KS

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.ks.tax.income.credits
        # determine if tax unit is eligible for credit
        person = tax_unit.members
        # ... any child dependents?
        count_all_dependents = tax_unit("tax_unit_dependents", period)
        child_age = p.food_sales_tax.child_age
        is_child = person("age", period) < child_age
        is_child_dependent = person("is_tax_unit_dependent", period) & is_child
        count_child_dependents = tax_unit.sum(is_child_dependent)
        has_eligible_child = count_child_dependents > 0
        # ... any elderly adults?
        min_adult_age = p.food_sales_tax.min_adult_age
        elderly_head = tax_unit("age_head", period) >= min_adult_age
        elderly_spouse = tax_unit("age_spouse", period) >= min_adult_age
        eligible_age = elderly_head | elderly_spouse
        # ... any adult disabilities?
        eligible_blind_disabled = (
            tax_unit("blind_head", period)
            | tax_unit("blind_spouse", period)
            | tax_unit("head_is_disabled", period)
            | tax_unit("spouse_is_disabled", period)
        )
        # ... any eligibility for credit?
        eligible_unit = (
            has_eligible_child | eligible_age | eligible_blind_disabled
        )
        # determine if income eligible for credit
        fagi = tax_unit("adjusted_gross_income", period)
        eligible_income = fagi <= p.food_sales_tax.agi_limit
        # compute credit amount
        eligible = eligible_unit & eligible_income
        exemptions = tax_unit("ks_count_exemptions", period)
        old_dependents = count_all_dependents - count_child_dependents
        fstc_exemptions = max_(0, exemptions - old_dependents)
        return eligible * fstc_exemptions * p.food_sales_tax.amount



from policyengine_us.model_api import *


class wi_unemployment_compensation_subtraction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Wisconsin unemployment compensation subtraction from federal AGI"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.revenue.wi.gov/TaxForms2021/2021-ScheduleSB-inst.pdf#page=2"
        "https://www.revenue.wi.gov/TaxForms2022/2022-ScheduleSB-Inst.pdf#page=1"
    )
    defined_for = StateCode.WI

    def formula(tax_unit, period, parameters):
        us_tuc = add(tax_unit, period, ["taxable_unemployment_compensation"])
        us_agi = tax_unit("adjusted_gross_income", period)
        wi = parameters(period).gov.states.wi.tax.income
        p = wi.subtractions.unemployment_compensation.income_phase_out
        base_income = p.base[tax_unit("filing_status", period)]
        us_taxed_socsec = tax_unit("tax_unit_taxable_social_security", period)
        ucsub_income = base_income + us_taxed_socsec
        excess_inc = max_(0, us_agi - ucsub_income)
        tuc_limit = excess_inc * p.rate  # WI Unemploy Compen Worksheet, Line 8
        wi_tuc = min_(tuc_limit, us_tuc)
        return max_(0, us_tuc - wi_tuc)



from policyengine_us.model_api import *


class ca_tanf_financial_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "California CalWORKs Financial Eligibility"
    definition_period = YEAR
    defined_for = StateCode.CA

    def formula(spm_unit, period, parameters):
        tanf_enrolled = spm_unit("is_tanf_enrolled", period)
        applicant_test = spm_unit("ca_tanf_applicant_financial_test", period)
        recipient_test = spm_unit("ca_tanf_recipient_financial_test", period)

        return where(
            tanf_enrolled, recipient_test, applicant_test & recipient_test
        )



from policyengine_us.model_api import *


class sc_state_tax_addback(Variable):
    value_type = float
    entity = TaxUnit
    label = "South Carolina State Tax addback"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://dor.sc.gov/forms-site/Forms/SC1040_2022.pdf#page=2",
        "https://dor.sc.gov/forms-site/Forms/SC1040inst_2022.pdf#page=2",
        "https://www.scstatehouse.gov/code/t12c006.php",  # SECTION 12-6-1130 (2)
    )
    defined_for = StateCode.SC

    def formula(tax_unit, period, parameters):
        p_us = parameters(period).gov.irs.deductions

        us_itemizing = tax_unit("tax_unit_itemizes", period)
        standard_deduction = tax_unit("standard_deduction", period)
        filing_status = tax_unit("filing_status", period)
        eligible = filing_status != filing_status.possible_values.SEPARATE
        # line 1
        federal_itemized_deduction = (
            tax_unit("itemized_taxable_income_deductions", period)
            * us_itemizing
        )
        # line 2
        federal_standard_deduction = standard_deduction * eligible
        # line 3
        less_itm_amount = max_(
            0, federal_itemized_deduction - federal_standard_deduction
        )
        # line 4
        salt = tax_unit("state_and_local_sales_or_income_tax", period)
        # line 5
        real_estate_and_property_taxes = add(
            tax_unit, period, ["real_estate_taxes"]
        )
        less_income_amount = max_(
            0,
            p_us.itemized.salt_and_real_estate.cap[filing_status]
            - real_estate_and_property_taxes,
        )
        # compare line 3,4,5. get the minimum
        salt_or_income = min_(salt, less_income_amount)
        return min_(salt_or_income, less_itm_amount)



from policyengine_us.model_api import *


class is_tce_eligible(Variable):
    value_type = bool
    entity = Person
    label = "Eligible person for the Tax Counseling for the Elderly program"
    definition_period = YEAR
    reference = (
        "https://www.irs.gov/individuals/tax-counseling-for-the-elderly"
    )

    def formula(person, period, parameters):
        p = parameters(period).gov.irs.tce
        age = person("age", period)
        return age >= p.age_threshold



from policyengine_us.model_api import *


class de_cdcc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Delaware dependent care credit"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.DE

    def formula(tax_unit, period, parameters):
        # Delaware matches the federal credit taken
        expenses = tax_unit("cdcc", period)
        rate = parameters(period).gov.states.de.tax.income.credits.cdcc.match
        return expenses * rate



from policyengine_us.model_api import *

# FIXME: This is boilerplate taken directly from the reform code.


class co_family_affordability_credit(Variable):
    value_type = float
    entity = Person
    label = "Colorado Family Affordability Credit"
    unit = USD
    definition_period = YEAR
    reference = "https://leg.colorado.gov/bills/hb24-1311"
    defined_for = StateCode.CO

    def formula(person, period, parameters):
        age = person("age", period)
        dependent = person("is_child_dependent", period)
        p = parameters(
            period
        ).gov.states.co.tax.income.credits.family_affordability
        base_amount = p.amount * dependent
        agi = person.tax_unit("adjusted_gross_income", period)
        filing_status = person.tax_unit("filing_status", period)
        reduction_threshold = p.reduction.threshold[filing_status]
        excess = max_(agi - reduction_threshold, 0)
        increments = np.ceil(excess / p.reduction.increment)
        percent_reduction = min_(increments * p.reduction.rate, 1)
        age_multiplier = p.age_multiplier.calc(age)
        return base_amount * age_multiplier * (1 - percent_reduction)
